from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

import rich.repr
from rich.console import RenderableType
from rich.style import Style
from rich.text import Text, TextType

from textual import events
from textual.app import ComposeResult, RenderResult
from textual.await_complete import AwaitComplete
from textual.binding import Binding, BindingType
from textual.containers import Container, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.events import Mount
from textual.geometry import Offset
from textual.message import Message
from textual.reactive import reactive
from textual.renderables.bar import Bar
from textual.widget import Widget
from textual.widgets import Static


class Underline(Widget):
    """The animated underline beneath tabs."""

    DEFAULT_CSS = """
    Underline {
        width: 1fr;
        height: 1;
    }
    Underline > .underline--bar {
        background: $foreground 10%;
        color: $accent;
    }
    """

    COMPONENT_CLASSES = {"underline--bar"}
    """
    | Class | Description |
    | :- | :- |
    | `underline--bar` | Style of the bar (may be used to change the color). |
    """

    highlight_start = reactive(0)
    """First cell in highlight."""
    highlight_end = reactive(0)
    """Last cell (inclusive) in highlight."""
    show_highlight: reactive[bool] = reactive(True)
    """Flag to indicate if a highlight should be shown at all."""

    class Clicked(Message):
        """Inform ancestors the underline was clicked."""

        offset: Offset
        """The offset of the click, relative to the origin of the bar."""

        def __init__(self, offset: Offset) -> None:
            self.offset = offset
            super().__init__()

    @property
    def _highlight_range(self) -> tuple[int, int]:
        """Highlighted range for underline bar."""
        return (
            (self.highlight_start, self.highlight_end)
            if self.show_highlight
            else (0, 0)
        )

    def render(self) -> RenderResult:
        """Render the bar."""
        bar_style = self.get_component_rich_style("underline--bar")
        return Bar(
            highlight_range=self._highlight_range,
            highlight_style=Style.from_color(bar_style.color),
            background_style=Style.from_color(bar_style.bgcolor),
        )

    def _on_click(self, event: events.Click):
        """Catch clicks, so that the underline can activate the tabs."""
        event.stop()
        self.post_message(self.Clicked(event.screen_offset))


class Tab(Static):
    """A Widget to manage a single tab within a Tabs widget."""

    DEFAULT_CSS = """
    Tab {
        width: auto;
        height: 2;
        padding: 1 1 0 2;
        text-align: center;
        color: $text-disabled;
    }
    Tab.-active {
        text-style: bold;
        color: $text;
    }
    Tab:hover {
        text-style: bold;
    }
    Tab.-active:hover {
        color: $text;
    }
    Tab:disabled {
        color: $text-disabled;
        text-opacity: 50%;
    }
    Tab.-hidden {
        display: none;
    }
    """

    @dataclass
    class TabMessage(Message):
        """Tab-related messages.

        These are mostly intended for internal use when interacting with `Tabs`.
        """

        tab: Tab
        """The tab that is the object of this message."""

        @property
        def control(self) -> Tab:
            """The tab that is the object of this message.

            This is an alias for the attribute `tab` and is used by the
            [`on`][textual.on] decorator.
            """
            return self.tab

    class Clicked(TabMessage):
        """A tab was clicked."""

    class Disabled(TabMessage):
        """A tab was disabled."""

    class Enabled(TabMessage):
        """A tab was enabled."""

    class Relabelled(TabMessage):
        """A tab was relabelled."""

    def __init__(
        self,
        label: TextType,
        *,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise a Tab.

        Args:
            label: The label to use in the tab.
            id: Optional ID for the widget.
            classes: Space separated list of class names.
            disabled: Whether the tab is disabled or not.
        """
        super().__init__(id=id, classes=classes, disabled=disabled)
        self._label: Text
        # Setter takes Text or str
        self.label = label  # type: ignore[assignment]

    @property
    def label(self) -> Text:
        """The label for the tab."""
        return self._label

    @label.setter
    def label(self, label: TextType) -> None:
        self._label = Text.from_markup(label) if isinstance(label, str) else label
        self.update(self._label)

    def update(self, renderable: RenderableType = "") -> None:
        self.post_message(self.Relabelled(self))
        return super().update(renderable)

    @property
    def label_text(self) -> str:
        """Undecorated text of the label."""
        return self.label.plain

    def _on_click(self):
        """Inform the message that the tab was clicked."""
        self.post_message(self.Clicked(self))

    def _watch_disabled(self, disabled: bool) -> None:
        """Notify the parent `Tabs` that a tab was enabled/disabled."""
        self.post_message(self.Disabled(self) if disabled else self.Enabled(self))


class Tabs(Widget, can_focus=True):
    """A row of tabs."""

    DEFAULT_CSS = """
    Tabs {
        width: 100%;
        height: 3;
    }
    Tabs > #tabs-scroll {
        overflow: hidden;
    }
    Tabs #tabs-list {
       width: auto;
       min-height: 2;
    }
    Tabs #tabs-list-bar, Tabs #tabs-list {
        width: auto;
        height: auto;
        min-width: 100%;
        overflow: hidden hidden;
    }
    Tabs:focus .underline--bar {
        background: $foreground 20%;
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("left", "previous_tab", "Previous tab", show=False),
        Binding("right", "next_tab", "Next tab", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | left | Move to the previous tab. |
    | right | Move to the next tab. |
    """

    class TabError(Exception):
        """Exception raised when there is an error relating to tabs."""

    class TabMessage(Message):
        """Parent class for all messages that have to do with a specific tab."""

        ALLOW_SELECTOR_MATCH = {"tab"}
        """Additional message attributes that can be used with the [`on` decorator][textual.on]."""

        def __init__(self, tabs: Tabs, tab: Tab) -> None:
            """Initialize event.

            Args:
                tabs: The Tabs widget.
                tab: The tab that is the object of this message.
            """
            self.tabs: Tabs = tabs
            """The tabs widget containing the tab."""
            self.tab: Tab = tab
            """The tab that is the object of this message."""
            super().__init__()

        @property
        def control(self) -> Tabs:
            """The tabs widget containing the tab that is the object of this message.

            This is an alias for the attribute `tabs` and is used by the
            [`on`][textual.on] decorator.
            """
            return self.tabs

        def __rich_repr__(self) -> rich.repr.Result:
            yield self.tabs
            yield self.tab

    class TabActivated(TabMessage):
        """Sent when a new tab is activated."""

    class TabDisabled(TabMessage):
        """Sent when a tab is disabled."""

    class TabEnabled(TabMessage):
        """Sent when a tab is enabled."""

    class TabHidden(TabMessage):
        """Sent when a tab is hidden."""

    class TabShown(TabMessage):
        """Sent when a tab is shown."""

    class Cleared(Message):
        """Sent when there are no active tabs.

        This can occur when Tabs are cleared, if all tabs are hidden, or if the
        currently active tab is unset.
        """

        def __init__(self, tabs: Tabs) -> None:
            """Initialize the event.

            Args:
                tabs: The tabs widget.
            """
            self.tabs: Tabs = tabs
            """The tabs widget which was cleared."""
            super().__init__()

        @property
        def control(self) -> Tabs:
            """The tabs widget which was cleared.

            This is an alias for [`Cleared.tabs`][textual.widgets.Tabs.Cleared] which
            is used by the [`on`][textual.on] decorator.
            """
            return self.tabs

        def __rich_repr__(self) -> rich.repr.Result:
            yield self.tabs

    active: reactive[str] = reactive("", init=False)
    """The ID of the active tab, or empty string if none are active."""

    def __init__(
        self,
        *tabs: Tab | TextType,
        active: str | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        """Construct a Tabs widget.

        Args:
            *tabs: Positional argument should be explicit Tab objects, or a str or Text.
            active: ID of the tab which should be active on start.
            name: Optional name for the tabs widget.
            id: Optional ID for the widget.
            classes: Optional initial classes for the widget.
            disabled: Whether the widget is disabled or not.
        """
        self._tabs_counter = 0

        add_tabs = [
            (
                Tab(tab, id=f"tab-{self._new_tab_id}")
                if isinstance(tab, (str, Text))
                else self._auto_tab_id(tab)
            )
            for tab in tabs
        ]
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self._tabs = add_tabs
        self._first_active = active

    def _auto_tab_id(self, tab: Tab) -> Tab:
        """Set an automatic ID if not supplied."""
        if tab.id is None:
            tab.id = f"tab-{self._new_tab_id}"
        return tab

    @property
    def _new_tab_id(self) -> int:
        """Get the next tab id in a sequence."""
        self._tabs_counter += 1
        return self._tabs_counter

    @property
    def tab_count(self) -> int:
        """Total number of tabs."""
        return len(self.query("#tabs-list > Tab"))

    @property
    def _potentially_active_tabs(self) -> list[Tab]:
        """List of all tabs that could be active.

        This list is comprised of all tabs that are shown and enabled,
        plus the active tab in case it is disabled.
        """
        return [
            tab
            for tab in self.query("#tabs-list > Tab").results(Tab)
            if ((not tab.disabled or tab is self.active_tab) and tab.display)
        ]

    @property
    def _next_active(self) -> Tab | None:
        """Next tab to make active if the active tab is removed."""
        tabs = self._potentially_active_tabs
        if self.active_tab is None:
            return None
        try:
            active_index = tabs.index(self.active_tab)
        except ValueError:
            return None
        del tabs[active_index]
        try:
            return tabs[active_index]
        except IndexError:
            try:
                return tabs[active_index - 1]
            except IndexError:
                pass
        return None

    def add_tab(
        self,
        tab: Tab | str | Text,
        *,
        before: Tab | str | None = None,
        after: Tab | str | None = None,
    ) -> AwaitComplete:
        """Add a new tab to the end of the tab list.

        Args:
            tab: A new tab object, or a label (str or Text).
            before: Optional tab or tab ID to add the tab before.
            after: Optional tab or tab ID to add the tab after.

        Returns:
            An optionally awaitable object that waits for the tab to be mounted and
                internal state to be fully updated to reflect the new tab.

        Raises:
            Tabs.TabError: If there is a problem with the addition request.

        Note:
            Only one of `before` or `after` can be provided. If both are
            provided a `Tabs.TabError` will be raised.
        """

        if before and after:
            raise self.TabError("Unable to add a tab both before and after a tab")

        if isinstance(before, str):
            try:
                before = self.query_one(f"#tabs-list > #{before}", Tab)
            except NoMatches:
                raise self.TabError(
                    f"There is no tab with ID '{before}' to mount before"
                )
        elif isinstance(before, Tab) and self not in before.ancestors:
            raise self.TabError(
                "Request to add a tab before a tab that isn't part of this tab collection"
            )

        if isinstance(after, str):
            try:
                after = self.query_one(f"#tabs-list > #{after}", Tab)
            except NoMatches:
                raise self.TabError(f"There is no tab with ID '{after}' to mount after")
        elif isinstance(after, Tab) and self not in after.ancestors:
            raise self.TabError(
                "Request to add a tab after a tab that isn't part of this tab collection"
            )

        from_empty = self.tab_count == 0
        tab_widget = (
            Tab(tab, id=f"tab-{self._new_tab_id}")
            if isinstance(tab, (str, Text))
            else self._auto_tab_id(tab)
        )

        mount_await = self.query_one("#tabs-list").mount(
            tab_widget, before=before, after=after
        )

        if from_empty:
            tab_widget.add_class("-active")
            activated_message = self.TabActivated(self, tab_widget)

            async def refresh_active() -> None:
                """Wait for things to be mounted before highlighting."""
                await mount_await
                self.active = tab_widget.id or ""
                self._highlight_active(animate=False)
                self.post_message(activated_message)

            return AwaitComplete(refresh_active())
        elif before or after:

            async def refresh_active() -> None:
                await mount_await
                self._highlight_active(animate=False)

            return AwaitComplete(refresh_active())

        return AwaitComplete(mount_await())

    def clear(self) -> AwaitComplete:
        """Clear all the tabs.

        Returns:
            An awaitable object that waits for the tabs to be removed.
        """
        underline = self.query_one(Underline)
        underline.highlight_start = 0
        underline.highlight_end = 0
        self.call_after_refresh(self.post_message, self.Cleared(self))
        self.active = ""
        return AwaitComplete(self.query("#tabs-list > Tab").remove())

    def remove_tab(self, tab_or_id: Tab | str | None) -> AwaitComplete:
        """Remove a tab.

        Args:
            tab_or_id: The Tab to remove or its id.

        Returns:
            An optionally awaitable object that waits for the tab to be removed.
        """
        if not tab_or_id:
            return AwaitComplete()

        if isinstance(tab_or_id, Tab):
            remove_tab = tab_or_id
        else:
            try:
                remove_tab = self.query_one(f"#tabs-list > #{tab_or_id}", Tab)
            except NoMatches:
                return AwaitComplete()

        if remove_tab.has_class("-active"):
            next_tab = self._next_active
        else:
            next_tab = None

        async def do_remove() -> None:
            """Perform the remove after refresh so the underline bar gets new positions."""
            await remove_tab.remove()
            if next_tab is not None:
                self.active = next_tab.id or ""
            if not self.query("#tabs-list > Tab"):
                self.active = ""

        return AwaitComplete(do_remove())

    def validate_active(self, active: str) -> str:
        """Check id assigned to active attribute is a valid tab."""
        if active and not self.query(f"#tabs-list > #{active}"):
            raise ValueError(f"No Tab with id {active!r}")
        return active

    @property
    def active_tab(self) -> Tab | None:
        """The currently active tab, or None if there are no active tabs."""
        try:
            return self.query_one("#tabs-list Tab.-active", Tab)
        except NoMatches:
            return None

    def _on_mount(self, _: Mount) -> None:
        """Make the first tab active."""
        if self._first_active is not None:
            self.active = self._first_active
        if not self.active:
            try:
                tab = self.query("#tabs-list > Tab").first(Tab)
            except NoMatches:
                # Tabs are empty!
                return
            self.active = tab.id or ""

    def compose(self) -> ComposeResult:
        with Container(id="tabs-scroll"):
            with Vertical(id="tabs-list-bar"):
                with Horizontal(id="tabs-list"):
                    yield from self._tabs
                yield Underline()

    def watch_active(self, previously_active: str, active: str) -> None:
        """Handle a change to the active tab."""
        self.query("#tabs-list > Tab.-active").remove_class("-active")
        if active:
            try:
                active_tab = self.query_one(f"#tabs-list > #{active}", Tab)
            except NoMatches:
                return
            active_tab.add_class("-active")

            self._highlight_active(animate=previously_active != "")

            self._scroll_active_tab()
            self.post_message(self.TabActivated(self, active_tab))
        else:
            underline = self.query_one(Underline)
            underline.highlight_start = 0
            underline.highlight_end = 0
            self.post_message(self.Cleared(self))

    def _highlight_active(
        self,
        animate: bool = True,
    ) -> None:
        """Move the underline bar to under the active tab.

        Args:
            animate: Should the bar animate?
        """
        underline = self.query_one(Underline)
        try:
            _active_tab = self.query_one("#tabs-list > Tab.-active")
        except NoMatches:
            underline.show_highlight = False
            underline.highlight_start = 0
            underline.highlight_end = 0
        else:
            underline.show_highlight = True

            def move_underline(animate: bool) -> None:
                """Move the tab underline.

                Args:
                    animate: animate the underline to its new position.
                """
                try:
                    active_tab = self.query_one("#tabs-list > Tab.-active")
                except NoMatches:
                    pass
                else:
                    tab_region = active_tab.virtual_region.shrink(
                        active_tab.styles.gutter
                    )
                    start, end = tab_region.column_span
                    if animate:
                        underline.animate(
                            "highlight_start",
                            start,
                            duration=0.3,
                            level="basic",
                        )
                        underline.animate(
                            "highlight_end",
                            end,
                            duration=0.3,
                            level="basic",
                        )
                    else:
                        underline.highlight_start = start
                        underline.highlight_end = end

            if animate and self.app.animation_level != "none":
                self.set_timer(
                    0.02,
                    lambda: self.call_after_refresh(move_underline, True),
                )
            else:
                self.call_after_refresh(move_underline, False)

    async def _on_tab_clicked(self, event: Tab.Clicked) -> None:
        """Activate a tab that was clicked."""
        self.focus()
        event.stop()
        self._activate_tab(event.tab)

    def _activate_tab(self, tab: Tab) -> None:
        """Activate a tab.

        Args:
            tab: The Tab that was clicked.
        """
        self.query("#tabs-list Tab.-active").remove_class("-active")
        tab.add_class("-active")
        self.active = tab.id or ""

    def _on_underline_clicked(self, event: Underline.Clicked) -> None:
        """The underline was clicked.

        Activate the tab above to make a larger clickable area.

        Args:
            event: The Underline.Clicked event.
        """
        event.stop()
        offset = event.offset + (0, -1)
        self.focus()
        for tab in self.query(Tab):
            if offset in tab.region and not tab.disabled:
                self._activate_tab(tab)
                break

    def _scroll_active_tab(self) -> None:
        """Scroll the active tab into view."""
        if self.active_tab:
            try:
                self.query_one("#tabs-scroll").scroll_to_center(
                    self.active_tab, force=True
                )
            except NoMatches:
                pass

    def _on_resize(self):
        """Make the active tab visible on resize."""
        self._highlight_active(animate=False)
        self._scroll_active_tab()

    def action_next_tab(self) -> None:
        """Make the next tab active."""
        self._move_tab(+1)

    def action_previous_tab(self) -> None:
        """Make the previous tab active."""
        self._move_tab(-1)

    def _move_tab(self, direction: int) -> None:
        """Activate the next enabled tab in the given direction.

        Tab selection wraps around. If no tab is currently active, the "next"
        tab is set to be the first and the "previous" tab is the last one.

        Args:
            direction: +1 for the next tab, -1 for the previous.
        """
        active_tab = self.active_tab
        tabs = self._potentially_active_tabs
        if not tabs:
            return
        if not active_tab:
            self.active = tabs[0 if direction == 1 else -1].id or ""
            return
        tab_count = len(tabs)
        new_tab_index = (tabs.index(active_tab) + direction) % tab_count
        self.active = tabs[new_tab_index].id or ""

    def _on_tab_disabled(self, event: Tab.Disabled) -> None:
        """Re-post the disabled message."""
        event.stop()
        self.post_message(self.TabDisabled(self, event.tab))

    def _on_tab_enabled(self, event: Tab.Enabled) -> None:
        """Re-post the enabled message."""
        event.stop()
        self.post_message(self.TabEnabled(self, event.tab))

    def _on_tab_relabelled(self, event: Tab.Relabelled) -> None:
        """Redraw the highlight when tab is relabelled."""
        event.stop()
        self._highlight_active()

    def disable(self, tab_id: str) -> Tab:
        """Disable the indicated tab.

        Args:
            tab_id: The ID of the [`Tab`][textual.widgets.Tab] to disable.

        Returns:
            The [`Tab`][textual.widgets.Tab] that was targeted.

        Raises:
            TabError: If there are any issues with the request.
        """

        try:
            tab_to_disable = self.query_one(f"#tabs-list > Tab#{tab_id}", Tab)
        except NoMatches:
            raise self.TabError(
                f"There is no tab with ID {tab_id!r} to disable."
            ) from None

        tab_to_disable.disabled = True
        return tab_to_disable

    def enable(self, tab_id: str) -> Tab:
        """Enable the indicated tab.

        Args:
            tab_id: The ID of the [`Tab`][textual.widgets.Tab] to enable.

        Returns:
            The [`Tab`][textual.widgets.Tab] that was targeted.

        Raises:
            TabError: If there are any issues with the request.
        """

        try:
            tab_to_enable = self.query_one(f"#tabs-list > Tab#{tab_id}", Tab)
        except NoMatches:
            raise self.TabError(
                f"There is no tab with ID {tab_id!r} to enable."
            ) from None

        tab_to_enable.disabled = False
        return tab_to_enable

    def hide(self, tab_id: str) -> Tab:
        """Hide the indicated tab.

        Args:
            tab_id: The ID of the [`Tab`][textual.widgets.Tab] to hide.

        Returns:
            The [`Tab`][textual.widgets.Tab] that was targeted.

        Raises:
            TabError: If there are any issues with the request.
        """

        try:
            tab_to_hide = self.query_one(f"#tabs-list > Tab#{tab_id}", Tab)
        except NoMatches:
            raise self.TabError(f"There is no tab with ID {tab_id!r} to hide.")

        if tab_to_hide.has_class("-active"):
            next_tab = self._next_active
            self.active = next_tab.id or "" if next_tab else ""
        tab_to_hide.add_class("-hidden")
        self.post_message(self.TabHidden(self, tab_to_hide).set_sender(self))
        self.call_after_refresh(self._highlight_active)
        return tab_to_hide

    def show(self, tab_id: str) -> Tab:
        """Show the indicated tab.

        Args:
            tab_id: The ID of the [`Tab`][textual.widgets.Tab] to show.

        Returns:
            The [`Tab`][textual.widgets.Tab] that was targeted.

        Raises:
            TabError: If there are any issues with the request.
        """

        try:
            tab_to_show = self.query_one(f"#tabs-list > Tab#{tab_id}", Tab)
        except NoMatches:
            raise self.TabError(f"There is no tab with ID {tab_id!r} to show.")

        tab_to_show.remove_class("-hidden")
        self.post_message(self.TabShown(self, tab_to_show).set_sender(self))
        if not self.active:
            self._activate_tab(tab_to_show)
        self.call_after_refresh(self._highlight_active)
        return tab_to_show



from __future__ import annotations

import asyncio
import os
import selectors
import signal
import sys
import termios
import tty
from codecs import getincrementaldecoder
from threading import Event, Thread
from typing import TYPE_CHECKING, Any

import rich.repr

from textual import events
from textual._loop import loop_last
from textual._parser import ParseError
from textual._xterm_parser import XTermParser
from textual.driver import Driver
from textual.geometry import Size

if TYPE_CHECKING:
    from textual.app import App


@rich.repr.auto(angular=True)
class LinuxInlineDriver(Driver):
    def __init__(
        self,
        app: App,
        *,
        debug: bool = False,
        mouse: bool = True,
        size: tuple[int, int] | None = None,
    ):
        super().__init__(app, debug=debug, mouse=mouse, size=size)
        self._file = sys.__stderr__
        self.fileno = sys.__stdin__.fileno()
        self.attrs_before: list[Any] | None = None
        self.exit_event = Event()

    def __rich_repr__(self) -> rich.repr.Result:
        yield self._app

    @property
    def is_inline(self) -> bool:
        return True

    def _enable_bracketed_paste(self) -> None:
        """Enable bracketed paste mode."""
        self.write("\x1b[?2004h")

    def _disable_bracketed_paste(self) -> None:
        """Disable bracketed paste mode."""
        self.write("\x1b[?2004l")

    def _get_terminal_size(self) -> tuple[int, int]:
        """Detect the terminal size.

        Returns:
            The size of the terminal as a tuple of (WIDTH, HEIGHT).
        """
        width: int | None = 80
        height: int | None = 25
        import shutil

        try:
            width, height = shutil.get_terminal_size()
        except (AttributeError, ValueError, OSError):
            try:
                width, height = shutil.get_terminal_size()
            except (AttributeError, ValueError, OSError):
                pass
        width = width or 80
        height = height or 25
        return width, height

    def _enable_mouse_support(self) -> None:
        """Enable reporting of mouse events."""
        if not self._mouse:
            return
        write = self.write
        write("\x1b[?1000h")  # SET_VT200_MOUSE
        write("\x1b[?1003h")  # SET_ANY_EVENT_MOUSE
        write("\x1b[?1015h")  # SET_VT200_HIGHLIGHT_MOUSE
        write("\x1b[?1006h")  # SET_SGR_EXT_MODE_MOUSE

        # write("\x1b[?1007h")
        self.flush()

    def _disable_mouse_support(self) -> None:
        """Disable reporting of mouse events."""
        if not self._mouse:
            return
        write = self.write
        write("\x1b[?1000l")  #
        write("\x1b[?1003l")  #
        write("\x1b[?1015l")
        write("\x1b[?1006l")
        self.flush()

    def write(self, data: str) -> None:
        self._file.write(data)

    def _run_input_thread(self) -> None:
        """
        Key thread target that wraps run_input_thread() to die gracefully if it raises
        an exception
        """
        try:
            self.run_input_thread()
        except BaseException:
            import rich.traceback

            self._app.call_later(
                self._app.panic,
                rich.traceback.Traceback(),
            )

    def run_input_thread(self) -> None:
        """Wait for input and dispatch events."""
        selector = selectors.SelectSelector()
        selector.register(self.fileno, selectors.EVENT_READ)

        fileno = self.fileno
        EVENT_READ = selectors.EVENT_READ

        parser = XTermParser(self._debug)
        feed = parser.feed
        tick = parser.tick

        utf8_decoder = getincrementaldecoder("utf-8")().decode
        decode = utf8_decoder
        read = os.read

        def process_selector_events(
            selector_events: list[tuple[selectors.SelectorKey, int]],
            final: bool = False,
        ) -> None:
            """Process events from selector.

            Args:
                selector_events: List of selector events.
                final: True if this is the last call.

            """
            for last, (_selector_key, mask) in loop_last(selector_events):
                if mask & EVENT_READ:
                    unicode_data = decode(read(fileno, 1024 * 4), final=final and last)
                    if not unicode_data:
                        # This can occur if the stdin is piped
                        break
                    for event in feed(unicode_data):
                        if isinstance(event, events.CursorPosition):
                            self.cursor_origin = (event.x, event.y)
                        else:
                            self.process_event(event)
            for event in tick():
                if isinstance(event, events.CursorPosition):
                    self.cursor_origin = (event.x, event.y)
                else:
                    self.process_event(event)

        try:
            while not self.exit_event.is_set():
                process_selector_events(selector.select(0.1))
            selector.unregister(self.fileno)
            process_selector_events(selector.select(0.1), final=True)

        finally:
            selector.close()
            try:
                for event in feed(""):
                    pass
            except ParseError:
                pass

    def start_application_mode(self) -> None:
        loop = asyncio.get_running_loop()

        def send_size_event(clear: bool = False) -> None:
            """Send the resize event, optionally clearing the screen.

            Args:
                clear: Clear the screen.
            """
            terminal_size = self._get_terminal_size()
            width, height = terminal_size
            textual_size = Size(width, height)
            event = events.Resize(textual_size, textual_size)

            async def update_size() -> None:
                """Update the screen size."""
                if clear:
                    self.write("\x1b[2J")
                await self._app._post_message(event)

            asyncio.run_coroutine_threadsafe(
                update_size(),
                loop=loop,
            )

        def on_terminal_resize(signum, stack) -> None:
            send_size_event(clear=True)

        signal.signal(signal.SIGWINCH, on_terminal_resize)

        self.write("\x1b[?25l")  # Hide cursor
        self.write("\033[?1004h")  # Enable FocusIn/FocusOut.

        self.write("\x1b[>1u")  # https://sw.kovidgoyal.net/kitty/keyboard-protocol/
        # Disambiguate escape codes https://sw.kovidgoyal.net/kitty/keyboard-protocol/#progressive-enhancement
        self.write("\x1b[=1;u")
        self.flush()

        self._enable_mouse_support()
        self.write("\n" * self._app.INLINE_PADDING)
        self.flush()
        try:
            self.attrs_before = termios.tcgetattr(self.fileno)
        except termios.error:
            # Ignore attribute errors.
            self.attrs_before = None

        try:
            newattr = termios.tcgetattr(self.fileno)
        except termios.error:
            pass
        else:
            newattr[tty.LFLAG] = self._patch_lflag(newattr[tty.LFLAG])
            newattr[tty.IFLAG] = self._patch_iflag(newattr[tty.IFLAG])

            # VMIN defines the number of characters read at a time in
            # non-canonical mode. It seems to default to 1 on Linux, but on
            # Solaris and derived operating systems it defaults to 4. (This is
            # because the VMIN slot is the same as the VEOF slot, which
            # defaults to ASCII EOT = Ctrl-D = 4.)
            newattr[tty.CC][termios.VMIN] = 1

            termios.tcsetattr(self.fileno, termios.TCSANOW, newattr)

        self._key_thread = Thread(target=self._run_input_thread)
        send_size_event()
        self._key_thread.start()
        self._request_terminal_sync_mode_support()
        self._enable_bracketed_paste()

    def _request_terminal_sync_mode_support(self) -> None:
        """Writes an escape sequence to query the terminal support for the sync protocol."""
        # Terminals should ignore this sequence if not supported.
        # Apple terminal doesn't, and writes a single 'p' in to the terminal,
        # so we will make a special case for Apple terminal (which doesn't support sync anyway).
        if os.environ.get("TERM_PROGRAM", "") != "Apple_Terminal":
            self.write("\033[?2026$p")
            self.flush()

    @classmethod
    def _patch_lflag(cls, attrs: int) -> int:
        """Patch termios lflag.

        Args:
            attributes: New set attributes.

        Returns:
            New lflag.

        """
        # if TEXTUAL_ALLOW_SIGNALS env var is set, then allow Ctrl+C to send signals
        ISIG = 0 if os.environ.get("TEXTUAL_ALLOW_SIGNALS") else termios.ISIG

        return attrs & ~(termios.ECHO | termios.ICANON | termios.IEXTEN | ISIG)

    @classmethod
    def _patch_iflag(cls, attrs: int) -> int:
        return attrs & ~(
            # Disable XON/XOFF flow control on output and input.
            # (Don't capture Ctrl-S and Ctrl-Q.)
            # Like executing: "stty -ixon."
            termios.IXON
            | termios.IXOFF
            |
            # Don't translate carriage return into newline on input.
            termios.ICRNL
            | termios.INLCR
            | termios.IGNCR
        )

    def disable_input(self) -> None:
        """Disable further input."""
        try:
            if not self.exit_event.is_set():
                signal.signal(signal.SIGWINCH, signal.SIG_DFL)
                self._disable_mouse_support()
                self.exit_event.set()
                if self._key_thread is not None:
                    self._key_thread.join()
                self.exit_event.clear()
                try:
                    termios.tcflush(self.fileno, termios.TCIFLUSH)
                except termios.error:
                    pass

        except Exception as error:
            # TODO: log this
            pass

    def flush(self):
        """Flush any buffered data."""
        self._file.flush()

    def stop_application_mode(self) -> None:
        """Stop application mode, restore state."""
        self._disable_bracketed_paste()
        self.disable_input()
        self.write("\x1b[<u")  # Disable kitty protocol
        self.write("\x1b[J")

        if self.attrs_before is not None:
            try:
                termios.tcsetattr(self.fileno, termios.TCSANOW, self.attrs_before)
            except termios.error:
                pass

            self.write("\x1b[?25h")  # Show cursor
            self.write("\033[?1004l")  # Disable FocusIn/FocusOut.

        self.flush()



from textual.app import App, ComposeResult
from textual.widgets import Label, Button


class QuestionApp(App[str]):
    def compose(self) -> ComposeResult:
        yield Label("Do you love Textual?")
        yield Button("Yes", id="yes", variant="primary")
        yield Button("No", id="no", variant="error")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.exit(event.button.id)


if __name__ == "__main__":
    app = QuestionApp()
    reply = app.run()
    print(reply)



from textual.app import App, ComposeResult
from textual.widgets import Button, Input


class PreventApp(App):
    """Demonstrates `prevent` context manager."""

    def compose(self) -> ComposeResult:
        yield Input()
        yield Button("Clear", id="clear")

    def on_button_pressed(self) -> None:
        """Clear the text input."""
        input = self.query_one(Input)
        with input.prevent(Input.Changed):  # (1)!
            input.value = ""

    def on_input_changed(self) -> None:
        """Called as the user types."""
        self.bell()  # (2)!


if __name__ == "__main__":
    app = PreventApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Header, Footer, Label


class VerticalRemoveApp(App[None]):
    CSS = """
    Vertical {
        border: round green;
        height: auto;
    }

    Label {
        border: round yellow;
        background: red;
        color: yellow;
    }
    """
    BINDINGS = [
        ("a", "add", "Add"),
        ("d", "del", "Delete"),
    ]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Vertical()
        yield Footer()

    def action_add(self) -> None:
        self.query_one(Vertical).mount(Label("This is a test label"))

    def action_del(self) -> None:
        if self.query_one(Vertical).children:
            self.query_one(Vertical).children[-1].remove()


if __name__ == "__main__":
    VerticalRemoveApp().run()



"""

Functions and classes to manage terminal geometry (anything involving coordinates or dimensions).
"""

from __future__ import annotations

from functools import lru_cache
from operator import attrgetter, itemgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    NamedTuple,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Final

if TYPE_CHECKING:
    from typing_extensions import TypeAlias


SpacingDimensions: TypeAlias = Union[
    int, Tuple[int], Tuple[int, int], Tuple[int, int, int, int]
]
"""The valid ways in which you can specify spacing."""

T = TypeVar("T", int, float)


def clamp(value: T, minimum: T, maximum: T) -> T:
    """Restrict a value to a given range.

    If `value` is less than the minimum, return the minimum.
    If `value` is greater than the maximum, return the maximum.
    Otherwise, return `value`.

    The `minimum` and `maximum` arguments values may be given in reverse order.

    Args:
        value: A value.
        minimum: Minimum value.
        maximum: Maximum value.

    Returns:
        New value that is not less than the minimum or greater than the maximum.
    """
    if minimum > maximum:
        # It is common for the min and max to be in non-intuitive order.
        # Rather than force the caller to get it right, it is simpler to handle it here.
        if value < maximum:
            return maximum
        if value > minimum:
            return minimum
        return value
    else:
        if value < minimum:
            return minimum
        if value > maximum:
            return maximum
        return value


class Offset(NamedTuple):
    """A cell offset defined by x and y coordinates.

    Offsets are typically relative to the top left of the terminal or other container.

    Textual prefers the names `x` and `y`, but you could consider `x` to be the _column_ and `y` to be the _row_.

    Offsets support addition, subtraction, multiplication, and negation.

    Example:
        ```python
        >>> from textual.geometry import Offset
        >>> offset = Offset(3, 2)
        >>> offset
        Offset(x=3, y=2)
        >>> offset += Offset(10, 0)
        >>> offset
        Offset(x=13, y=2)
        >>> -offset
        Offset(x=-13, y=-2)
        ```
    """

    x: int = 0
    """Offset in the x-axis (horizontal)"""
    y: int = 0
    """Offset in the y-axis (vertical)"""

    @property
    def is_origin(self) -> bool:
        """Is the offset at (0, 0)?"""
        return self == (0, 0)

    @property
    def clamped(self) -> Offset:
        """This offset with `x` and `y` restricted to values above zero."""
        x, y = self
        return Offset(0 if x < 0 else x, 0 if y < 0 else y)

    def __bool__(self) -> bool:
        return self != (0, 0)

    def __add__(self, other: object) -> Offset:
        if isinstance(other, tuple):
            _x, _y = self
            x, y = other
            return Offset(_x + x, _y + y)
        return NotImplemented

    def __sub__(self, other: object) -> Offset:
        if isinstance(other, tuple):
            _x, _y = self
            x, y = other
            return Offset(_x - x, _y - y)
        return NotImplemented

    def __mul__(self, other: object) -> Offset:
        if isinstance(other, (float, int)):
            x, y = self
            return Offset(int(x * other), int(y * other))
        return NotImplemented

    def __neg__(self) -> Offset:
        x, y = self
        return Offset(-x, -y)

    def blend(self, destination: Offset, factor: float) -> Offset:
        """Calculate a new offset on a line between this offset and a destination offset.

        Args:
            destination: Point where factor would be 1.0.
            factor: A value between 0 and 1.0.

        Returns:
            A new point on a line between self and destination.
        """
        x1, y1 = self
        x2, y2 = destination
        return Offset(
            int(x1 + (x2 - x1) * factor),
            int(y1 + (y2 - y1) * factor),
        )

    def get_distance_to(self, other: Offset) -> float:
        """Get the distance to another offset.

        Args:
            other: An offset.

        Returns:
            Distance to other offset.
        """
        x1, y1 = self
        x2, y2 = other
        distance: float = ((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1)) ** 0.5
        return distance

    def clamp(self, width: int, height: int) -> Offset:
        """Clamp the offset to fit within a rectangle of width x height.

        Args:
            width: Width to clamp.
            height: Height to clamp.

        Returns:
            A new offset.
        """
        x, y = self
        return Offset(clamp(x, 0, width - 1), clamp(y, 0, height - 1))


class Size(NamedTuple):
    """The dimensions (width and height) of a rectangular region.

    Example:
        ```python
        >>> from textual.geometry import Size
        >>> size = Size(2, 3)
        >>> size
        Size(width=2, height=3)
        >>> size.area
        6
        >>> size + Size(10, 20)
        Size(width=12, height=23)
        ```
    """

    width: int = 0
    """The width in cells."""

    height: int = 0
    """The height in cells."""

    def __bool__(self) -> bool:
        """A Size is Falsy if it has area 0."""
        return self.width * self.height != 0

    @property
    def area(self) -> int:
        """The area occupied by a region of this size."""
        return self.width * self.height

    @property
    def region(self) -> Region:
        """A region of the same size, at the origin."""
        width, height = self
        return Region(0, 0, width, height)

    @property
    def line_range(self) -> range:
        """A range object that covers values between 0 and `height`."""
        return range(self.height)

    def with_width(self, width: int) -> Size:
        """Get a new Size with just the width changed.

        Args:
            width: New width.

        Returns:
            New Size instance.
        """
        return Size(width, self.height)

    def with_height(self, height: int) -> Size:
        """Get a new Size with just the height changed.

        Args:
            height: New height.

        Returns:
            New Size instance.
        """
        return Size(self.width, height)

    def __add__(self, other: object) -> Size:
        if isinstance(other, tuple):
            width, height = self
            width2, height2 = other
            return Size(max(0, width + width2), max(0, height + height2))
        return NotImplemented

    def __sub__(self, other: object) -> Size:
        if isinstance(other, tuple):
            width, height = self
            width2, height2 = other
            return Size(max(0, width - width2), max(0, height - height2))
        return NotImplemented

    def contains(self, x: int, y: int) -> bool:
        """Check if a point is in area defined by the size.

        Args:
            x: X coordinate.
            y: Y coordinate.

        Returns:
            True if the point is within the region.
        """
        width, height = self
        return width > x >= 0 and height > y >= 0

    def contains_point(self, point: tuple[int, int]) -> bool:
        """Check if a point is in the area defined by the size.

        Args:
            point: A tuple of x and y coordinates.

        Returns:
            True if the point is within the region.
        """
        x, y = point
        width, height = self
        return width > x >= 0 and height > y >= 0

    def __contains__(self, other: Any) -> bool:
        try:
            x: int
            y: int
            x, y = other
        except Exception:
            raise TypeError(
                "Dimensions.__contains__ requires an iterable of two integers"
            )
        width, height = self
        return width > x >= 0 and height > y >= 0

    def clamp_offset(self, offset: Offset) -> Offset:
        """Clamp an offset to fit within the width x height.

        Args:
            offset: An offset.

        Returns:
            A new offset that will fit inside the dimensions defined in the Size.
        """
        return offset.clamp(self.width, self.height)


class Region(NamedTuple):
    """Defines a rectangular region.

    A Region consists of a coordinate (x and y) and dimensions (width and height).

    ```
      (x, y)
        ┌────────────────────┐ ▲
        │                    │ │
        │                    │ │
        │                    │ height
        │                    │ │
        │                    │ │
        └────────────────────┘ ▼
        ◀─────── width ──────▶
    ```

    Example:
        ```python
        >>> from textual.geometry import Region
        >>> region = Region(4, 5, 20, 10)
        >>> region
        Region(x=4, y=5, width=20, height=10)
        >>> region.area
        200
        >>> region.size
        Size(width=20, height=10)
        >>> region.offset
        Offset(x=4, y=5)
        >>> region.contains(1, 2)
        False
        >>> region.contains(10, 8)
        True
        ```
    """

    x: int = 0
    """Offset in the x-axis (horizontal)."""
    y: int = 0
    """Offset in the y-axis (vertical)."""
    width: int = 0
    """The width of the region."""
    height: int = 0
    """The height of the region."""

    @classmethod
    def from_union(cls, regions: Collection[Region]) -> Region:
        """Create a Region from the union of other regions.

        Args:
            regions: One or more regions.

        Returns:
            A Region that encloses all other regions.
        """
        if not regions:
            raise ValueError("At least one region expected")
        min_x = min(regions, key=itemgetter(0)).x
        max_x = max(regions, key=attrgetter("right")).right
        min_y = min(regions, key=itemgetter(1)).y
        max_y = max(regions, key=attrgetter("bottom")).bottom
        return cls(min_x, min_y, max_x - min_x, max_y - min_y)

    @classmethod
    def from_corners(cls, x1: int, y1: int, x2: int, y2: int) -> Region:
        """Construct a Region form the top left and bottom right corners.

        Args:
            x1: Top left x.
            y1: Top left y.
            x2: Bottom right x.
            y2: Bottom right y.

        Returns:
            A new region.
        """
        return cls(x1, y1, x2 - x1, y2 - y1)

    @classmethod
    def from_offset(cls, offset: tuple[int, int], size: tuple[int, int]) -> Region:
        """Create a region from offset and size.

        Args:
            offset: Offset (top left point).
            size: Dimensions of region.

        Returns:
            A region instance.
        """
        x, y = offset
        width, height = size
        return cls(x, y, width, height)

    @classmethod
    def get_scroll_to_visible(
        cls, window_region: Region, region: Region, *, top: bool = False
    ) -> Offset:
        """Calculate the smallest offset required to translate a window so that it contains
        another region.

        This method is used to calculate the required offset to scroll something in to view.

        Args:
            window_region: The window region.
            region: The region to move inside the window.
            top: Get offset to top of window.

        Returns:
            An offset required to add to region to move it inside window_region.
        """

        if region in window_region and not top:
            # Region is already inside the window, so no need to move it.
            return NULL_OFFSET

        window_left, window_top, window_right, window_bottom = window_region.corners
        region = region.crop_size(window_region.size)
        left, top_, right, bottom = region.corners
        delta_x = delta_y = 0

        if not (
            (window_right > left >= window_left)
            and (window_right > right >= window_left)
        ):
            # The region does not fit
            # The window needs to scroll on the X axis to bring region in to view
            delta_x = min(
                left - window_left,
                left - (window_right - region.width),
                key=abs,
            )

        if top:
            delta_y = top_ - window_top

        elif not (
            (window_bottom > top_ >= window_top)
            and (window_bottom > bottom >= window_top)
        ):
            # The window needs to scroll on the Y axis to bring region in to view
            delta_y = min(
                top_ - window_top,
                top_ - (window_bottom - region.height),
                key=abs,
            )
        return Offset(delta_x, delta_y)

    def __bool__(self) -> bool:
        """A Region is considered False when it has no area."""
        _, _, width, height = self
        return width * height > 0

    @property
    def column_span(self) -> tuple[int, int]:
        """A pair of integers for the start and end columns (x coordinates) in this region.

        The end value is *exclusive*.
        """
        return (self.x, self.x + self.width)

    @property
    def line_span(self) -> tuple[int, int]:
        """A pair of integers for the start and end lines (y coordinates) in this region.

        The end value is *exclusive*.
        """
        return (self.y, self.y + self.height)

    @property
    def right(self) -> int:
        """Maximum X value (non inclusive)."""
        return self.x + self.width

    @property
    def bottom(self) -> int:
        """Maximum Y value (non inclusive)."""
        return self.y + self.height

    @property
    def area(self) -> int:
        """The area under the region."""
        return self.width * self.height

    @property
    def offset(self) -> Offset:
        """The top left corner of the region.

        Returns:
            An offset.
        """
        return Offset(*self[:2])

    @property
    def center(self) -> tuple[float, float]:
        """The center of the region.

        Note, that this does *not* return an `Offset`, because the center may not be an integer coordinate.

        Returns:
            Tuple of floats.
        """
        x, y, width, height = self
        return (x + width / 2.0, y + height / 2.0)

    @property
    def bottom_left(self) -> Offset:
        """Bottom left offset of the region.

        Returns:
            An offset.
        """
        x, y, _width, height = self
        return Offset(x, y + height)

    @property
    def top_right(self) -> Offset:
        """Top right offset of the region.

        Returns:
            An offset.
        """
        x, y, width, _height = self
        return Offset(x + width, y)

    @property
    def bottom_right(self) -> Offset:
        """Bottom right offset of the region.

        Returns:
            An offset.
        """
        x, y, width, height = self
        return Offset(x + width, y + height)

    @property
    def size(self) -> Size:
        """Get the size of the region."""
        return Size(*self[2:])

    @property
    def corners(self) -> tuple[int, int, int, int]:
        """The top left and bottom right coordinates as a tuple of four integers."""
        x, y, width, height = self
        return x, y, x + width, y + height

    @property
    def column_range(self) -> range:
        """A range object for X coordinates."""
        return range(self.x, self.x + self.width)

    @property
    def line_range(self) -> range:
        """A range object for Y coordinates."""
        return range(self.y, self.y + self.height)

    @property
    def reset_offset(self) -> Region:
        """An region of the same size at (0, 0).

        Returns:
            A region at the origin.
        """
        _, _, width, height = self
        return Region(0, 0, width, height)

    def __add__(self, other: object) -> Region:
        if isinstance(other, tuple):
            ox, oy = other
            x, y, width, height = self
            return Region(x + ox, y + oy, width, height)
        return NotImplemented

    def __sub__(self, other: object) -> Region:
        if isinstance(other, tuple):
            ox, oy = other
            x, y, width, height = self
            return Region(x - ox, y - oy, width, height)
        return NotImplemented

    def get_spacing_between(self, region: Region) -> Spacing:
        """Get spacing between two regions.

        Args:
            region: Another region.

        Returns:
            Spacing that if subtracted from `self` produces `region`.
        """
        return Spacing(
            region.y - self.y,
            self.right - region.right,
            self.bottom - region.bottom,
            region.x - self.x,
        )

    def at_offset(self, offset: tuple[int, int]) -> Region:
        """Get a new Region with the same size at a given offset.

        Args:
            offset: An offset.

        Returns:
            New Region with adjusted offset.
        """
        x, y = offset
        _x, _y, width, height = self
        return Region(x, y, width, height)

    def crop_size(self, size: tuple[int, int]) -> Region:
        """Get a region with the same offset, with a size no larger than `size`.

        Args:
            size: Maximum width and height (WIDTH, HEIGHT).

        Returns:
            New region that could fit within `size`.
        """
        x, y, width1, height1 = self
        width2, height2 = size
        return Region(x, y, min(width1, width2), min(height1, height2))

    def expand(self, size: tuple[int, int]) -> Region:
        """Increase the size of the region by adding a border.

        Args:
            size: Additional width and height.

        Returns:
            A new region.
        """
        expand_width, expand_height = size
        x, y, width, height = self
        return Region(
            x - expand_width,
            y - expand_height,
            width + expand_width * 2,
            height + expand_height * 2,
        )

    @lru_cache(maxsize=1024)
    def overlaps(self, other: Region) -> bool:
        """Check if another region overlaps this region.

        Args:
            other: A Region.

        Returns:
            True if other region shares any cells with this region.
        """
        x, y, x2, y2 = self.corners
        ox, oy, ox2, oy2 = other.corners

        return ((x2 > ox >= x) or (x2 > ox2 > x) or (ox < x and ox2 >= x2)) and (
            (y2 > oy >= y) or (y2 > oy2 > y) or (oy < y and oy2 >= y2)
        )

    def contains(self, x: int, y: int) -> bool:
        """Check if a point is in the region.

        Args:
            x: X coordinate.
            y: Y coordinate.

        Returns:
            True if the point is within the region.
        """
        self_x, self_y, width, height = self
        return (self_x + width > x >= self_x) and (self_y + height > y >= self_y)

    def contains_point(self, point: tuple[int, int]) -> bool:
        """Check if a point is in the region.

        Args:
            point: A tuple of x and y coordinates.

        Returns:
            True if the point is within the region.
        """
        x1, y1, x2, y2 = self.corners
        try:
            ox, oy = point
        except Exception:
            raise TypeError(f"a tuple of two integers is required, not {point!r}")
        return (x2 > ox >= x1) and (y2 > oy >= y1)

    @lru_cache(maxsize=1024)
    def contains_region(self, other: Region) -> bool:
        """Check if a region is entirely contained within this region.

        Args:
            other: A region.

        Returns:
            True if the other region fits perfectly within this region.
        """
        x1, y1, x2, y2 = self.corners
        ox, oy, ox2, oy2 = other.corners
        return (
            (x2 >= ox >= x1)
            and (y2 >= oy >= y1)
            and (x2 >= ox2 >= x1)
            and (y2 >= oy2 >= y1)
        )

    @lru_cache(maxsize=1024)
    def translate(self, offset: tuple[int, int]) -> Region:
        """Move the offset of the Region.

        Args:
            offset: Offset to add to region.

        Returns:
            A new region shifted by (x, y).
        """

        self_x, self_y, width, height = self
        offset_x, offset_y = offset
        return Region(self_x + offset_x, self_y + offset_y, width, height)

    @lru_cache(maxsize=4096)
    def __contains__(self, other: Any) -> bool:
        """Check if a point is in this region."""
        if isinstance(other, Region):
            return self.contains_region(other)
        else:
            try:
                return self.contains_point(other)
            except TypeError:
                return False

    def clip(self, width: int, height: int) -> Region:
        """Clip this region to fit within width, height.

        Args:
            width: Width of bounds.
            height: Height of bounds.

        Returns:
            Clipped region.
        """
        x1, y1, x2, y2 = self.corners

        _clamp = clamp
        new_region = Region.from_corners(
            _clamp(x1, 0, width),
            _clamp(y1, 0, height),
            _clamp(x2, 0, width),
            _clamp(y2, 0, height),
        )
        return new_region

    @lru_cache(maxsize=4096)
    def grow(self, margin: tuple[int, int, int, int]) -> Region:
        """Grow a region by adding spacing.

        Args:
            margin: Grow space by `(<top>, <right>, <bottom>, <left>)`.

        Returns:
            New region.
        """
        if not any(margin):
            return self
        top, right, bottom, left = margin
        x, y, width, height = self
        return Region(
            x=x - left,
            y=y - top,
            width=max(0, width + left + right),
            height=max(0, height + top + bottom),
        )

    @lru_cache(maxsize=4096)
    def shrink(self, margin: tuple[int, int, int, int]) -> Region:
        """Shrink a region by subtracting spacing.

        Args:
            margin: Shrink space by `(<top>, <right>, <bottom>, <left>)`.

        Returns:
            The new, smaller region.
        """
        if not any(margin):
            return self
        top, right, bottom, left = margin
        x, y, width, height = self
        return Region(
            x=x + left,
            y=y + top,
            width=max(0, width - (left + right)),
            height=max(0, height - (top + bottom)),
        )

    @lru_cache(maxsize=4096)
    def intersection(self, region: Region) -> Region:
        """Get the overlapping portion of the two regions.

        Args:
            region: A region that overlaps this region.

        Returns:
            A new region that covers when the two regions overlap.
        """
        # Unrolled because this method is used a lot
        x1, y1, w1, h1 = self
        cx1, cy1, w2, h2 = region
        x2 = x1 + w1
        y2 = y1 + h1
        cx2 = cx1 + w2
        cy2 = cy1 + h2

        rx1 = cx2 if x1 > cx2 else (cx1 if x1 < cx1 else x1)
        ry1 = cy2 if y1 > cy2 else (cy1 if y1 < cy1 else y1)
        rx2 = cx2 if x2 > cx2 else (cx1 if x2 < cx1 else x2)
        ry2 = cy2 if y2 > cy2 else (cy1 if y2 < cy1 else y2)

        return Region(rx1, ry1, rx2 - rx1, ry2 - ry1)

    @lru_cache(maxsize=4096)
    def union(self, region: Region) -> Region:
        """Get the smallest region that contains both regions.

        Args:
            region: Another region.

        Returns:
            An optimally sized region to cover both regions.
        """
        x1, y1, x2, y2 = self.corners
        ox1, oy1, ox2, oy2 = region.corners

        union_region = self.from_corners(
            min(x1, ox1), min(y1, oy1), max(x2, ox2), max(y2, oy2)
        )
        return union_region

    @lru_cache(maxsize=1024)
    def split(self, cut_x: int, cut_y: int) -> tuple[Region, Region, Region, Region]:
        """Split a region in to 4 from given x and y offsets (cuts).

        ```
                   cut_x ↓
                ┌────────┐ ┌───┐
                │        │ │   │
                │    0   │ │ 1 │
                │        │ │   │
        cut_y → └────────┘ └───┘
                ┌────────┐ ┌───┐
                │    2   │ │ 3 │
                └────────┘ └───┘
        ```

        Args:
            cut_x: Offset from self.x where the cut should be made. If negative, the cut
                is taken from the right edge.
            cut_y: Offset from self.y where the cut should be made. If negative, the cut
                is taken from the lower edge.

        Returns:
            Four new regions which add up to the original (self).
        """

        x, y, width, height = self
        if cut_x < 0:
            cut_x = width + cut_x
        if cut_y < 0:
            cut_y = height + cut_y

        _Region = Region
        return (
            _Region(x, y, cut_x, cut_y),
            _Region(x + cut_x, y, width - cut_x, cut_y),
            _Region(x, y + cut_y, cut_x, height - cut_y),
            _Region(x + cut_x, y + cut_y, width - cut_x, height - cut_y),
        )

    @lru_cache(maxsize=1024)
    def split_vertical(self, cut: int) -> tuple[Region, Region]:
        """Split a region in to two, from a given x offset.

        ```
                 cut ↓
            ┌────────┐┌───┐
            │    0   ││ 1 │
            │        ││   │
            └────────┘└───┘
        ```

        Args:
            cut: An offset from self.x where the cut should be made. If cut is negative,
                it is taken from the right edge.

        Returns:
            Two regions, which add up to the original (self).
        """

        x, y, width, height = self
        if cut < 0:
            cut = width + cut

        return (
            Region(x, y, cut, height),
            Region(x + cut, y, width - cut, height),
        )

    @lru_cache(maxsize=1024)
    def split_horizontal(self, cut: int) -> tuple[Region, Region]:
        """Split a region in to two, from a given y offset.

        ```
                    ┌─────────┐
                    │    0    │
                    │         │
            cut →   └─────────┘
                    ┌─────────┐
                    │    1    │
                    └─────────┘
        ```

        Args:
            cut: An offset from self.y where the cut should be made. May be negative,
                for the offset to start from the lower edge.

        Returns:
            Two regions, which add up to the original (self).
        """
        x, y, width, height = self
        if cut < 0:
            cut = height + cut

        return (
            Region(x, y, width, cut),
            Region(x, y + cut, width, height - cut),
        )

    def translate_inside(
        self, container: Region, x_axis: bool = True, y_axis: bool = True
    ) -> Region:
        """Translate this region, so it fits within a container.

        This will ensure that there is as little overlap as possible.
        The top left of the returned region is guaranteed to be within the container.

        ```
        ┌──────────────────┐         ┌──────────────────┐
        │    container     │         │    container     │
        │                  │         │    ┌─────────────┤
        │                  │   ──▶   │    │    return   │
        │       ┌──────────┴──┐      │    │             │
        │       │    self     │      │    │             │
        └───────┤             │      └────┴─────────────┘
                │             │
                └─────────────┘
        ```


        Args:
            container: A container region.
            x_axis: Allow translation of X axis.
            y_axis: Allow translation of Y axis.

        Returns:
            A new region with same dimensions that fits with inside container.
        """
        x1, y1, width1, height1 = container
        x2, y2, width2, height2 = self
        return Region(
            max(min(x2, x1 + width1 - width2), x1) if x_axis else x2,
            max(min(y2, y1 + height1 - height2), y1) if y_axis else y2,
            width2,
            height2,
        )

    def inflect(
        self, x_axis: int = +1, y_axis: int = +1, margin: Spacing | None = None
    ) -> Region:
        """Inflect a region around one or both axis.

        The `x_axis` and `y_axis` parameters define which direction to move the region.
        A positive value will move the region right or down, a negative value will move
        the region left or up. A value of `0` will leave that axis unmodified.

        ```
        ╔══════════╗    │
        ║          ║
        ║   Self   ║    │
        ║          ║
        ╚══════════╝    │

        ─ ─ ─ ─ ─ ─ ─ ─ ┼ ─ ─ ─ ─ ─ ─ ─ ─

                        │    ┌──────────┐
                             │          │
                        │    │  Result  │
                             │          │
                        │    └──────────┘
        ```

        Args:
            x_axis: +1 to inflect in the positive direction, -1 to inflect in the negative direction.
            y_axis: +1 to inflect in the positive direction, -1 to inflect in the negative direction.
            margin: Additional margin.

        Returns:
            A new region.
        """
        inflect_margin = NULL_SPACING if margin is None else margin
        x, y, width, height = self
        if x_axis:
            x += (width + inflect_margin.width) * x_axis
        if y_axis:
            y += (height + inflect_margin.height) * y_axis
        return Region(x, y, width, height)


class Spacing(NamedTuple):
    """Stores spacing around a widget, such as padding and border.

    Spacing is defined by four integers for the space at the top, right, bottom, and left of a region.

    ```
    ┌ ─ ─ ─ ─ ─ ─ ─▲─ ─ ─ ─ ─ ─ ─ ─ ┐
                   │ top
    │        ┏━━━━━▼━━━━━━┓         │
     ◀──────▶┃            ┃◀───────▶
    │  left  ┃            ┃ right   │
             ┃            ┃
    │        ┗━━━━━▲━━━━━━┛         │
                   │ bottom
    └ ─ ─ ─ ─ ─ ─ ─▼─ ─ ─ ─ ─ ─ ─ ─ ┘
    ```

    Example:
        ```python
        >>> from textual.geometry import Region, Spacing
        >>> region = Region(2, 3, 20, 10)
        >>> spacing = Spacing(1, 2, 3, 4)
        >>> region.grow(spacing)
        Region(x=-2, y=2, width=26, height=14)
        >>> region.shrink(spacing)
        Region(x=6, y=4, width=14, height=6)
        >>> spacing.css
        '1 2 3 4'
        ```
    """

    top: int = 0
    """Space from the top of a region."""
    right: int = 0
    """Space from the right of a region."""
    bottom: int = 0
    """Space from the bottom of a region."""
    left: int = 0
    """Space from the left of a region."""

    def __bool__(self) -> bool:
        return self != (0, 0, 0, 0)

    @property
    def width(self) -> int:
        """Total space in the x axis."""
        return self.left + self.right

    @property
    def height(self) -> int:
        """Total space in the y axis."""
        return self.top + self.bottom

    @property
    def top_left(self) -> tuple[int, int]:
        """A pair of integers for the left, and top space."""
        return (self.left, self.top)

    @property
    def bottom_right(self) -> tuple[int, int]:
        """A pair of integers for the right, and bottom space."""
        return (self.right, self.bottom)

    @property
    def totals(self) -> tuple[int, int]:
        """A pair of integers for the total horizontal and vertical space."""
        top, right, bottom, left = self
        return (left + right, top + bottom)

    @property
    def css(self) -> str:
        """A string containing the spacing in CSS format.

        For example: "1" or "2 4" or "4 2 8 2".
        """
        top, right, bottom, left = self
        if top == right == bottom == left:
            return f"{top}"
        if (top, right) == (bottom, left):
            return f"{top} {right}"
        else:
            return f"{top} {right} {bottom} {left}"

    @classmethod
    def unpack(cls, pad: SpacingDimensions) -> Spacing:
        """Unpack padding specified in CSS style.

        Args:
            pad: An integer, or tuple of 1, 2, or 4 integers.

        Raises:
            ValueError: If `pad` is an invalid value.

        Returns:
            New Spacing object.
        """
        if isinstance(pad, int):
            return cls(pad, pad, pad, pad)
        pad_len = len(pad)
        if pad_len == 1:
            _pad = pad[0]
            return cls(_pad, _pad, _pad, _pad)
        if pad_len == 2:
            pad_top, pad_right = cast(Tuple[int, int], pad)
            return cls(pad_top, pad_right, pad_top, pad_right)
        if pad_len == 4:
            top, right, bottom, left = cast(Tuple[int, int, int, int], pad)
            return cls(top, right, bottom, left)
        raise ValueError(
            f"1, 2 or 4 integers required for spacing properties; {pad_len} given"
        )

    @classmethod
    def vertical(cls, amount: int) -> Spacing:
        """Construct a Spacing with a given amount of spacing on vertical edges,
        and no horizontal spacing.

        Args:
            amount: The magnitude of spacing to apply to vertical edges.

        Returns:
            `Spacing(amount, 0, amount, 0)`
        """
        return Spacing(amount, 0, amount, 0)

    @classmethod
    def horizontal(cls, amount: int) -> Spacing:
        """Construct a Spacing with a given amount of spacing on horizontal edges,
        and no vertical spacing.

        Args:
            amount: The magnitude of spacing to apply to horizontal edges.

        Returns:
            `Spacing(0, amount, 0, amount)`
        """
        return Spacing(0, amount, 0, amount)

    @classmethod
    def all(cls, amount: int) -> Spacing:
        """Construct a Spacing with a given amount of spacing on all edges.

        Args:
            amount: The magnitude of spacing to apply to all edges.

        Returns:
            `Spacing(amount, amount, amount, amount)`
        """
        return Spacing(amount, amount, amount, amount)

    def __add__(self, other: object) -> Spacing:
        if isinstance(other, tuple):
            top1, right1, bottom1, left1 = self
            top2, right2, bottom2, left2 = other
            return Spacing(
                top1 + top2, right1 + right2, bottom1 + bottom2, left1 + left2
            )
        return NotImplemented

    def __sub__(self, other: object) -> Spacing:
        if isinstance(other, tuple):
            top1, right1, bottom1, left1 = self
            top2, right2, bottom2, left2 = other
            return Spacing(
                top1 - top2, right1 - right2, bottom1 - bottom2, left1 - left2
            )
        return NotImplemented

    def grow_maximum(self, other: Spacing) -> Spacing:
        """Grow spacing with a maximum.

        Args:
            other: Spacing object.

        Returns:
            New spacing where the values are maximum of the two values.
        """
        top, right, bottom, left = self
        other_top, other_right, other_bottom, other_left = other
        return Spacing(
            max(top, other_top),
            max(right, other_right),
            max(bottom, other_bottom),
            max(left, other_left),
        )


NULL_OFFSET: Final = Offset(0, 0)
"""An [offset][textual.geometry.Offset] constant for (0, 0)."""

NULL_REGION: Final = Region(0, 0, 0, 0)
"""A [Region][textual.geometry.Region] constant for a null region (at the origin, with both width and height set to zero)."""

NULL_SIZE: Final = Size(0, 0)
"""A [Size][textual.geometry.Size] constant for a null size (with zero area)."""

NULL_SPACING: Final = Spacing(0, 0, 0, 0)
"""A [Spacing][textual.geometry.Spacing] constant for no space."""



from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, OptionList


class OptionListApp(App[None]):
    CSS_PATH = "option_list.tcss"

    def compose(self) -> ComposeResult:
        yield Header()
        yield OptionList(
            "Aerilon",
            "Aquaria",
            "Canceron",
            "Caprica",
            "Gemenon",
            "Leonis",
            "Libran",
            "Picon",
            "Sagittaron",
            "Scorpia",
            "Tauron",
            "Virgon",
        )
        yield Footer()


if __name__ == "__main__":
    OptionListApp().run()



from textual._easing import DEFAULT_EASING, EASING
from textual.app import App, ComposeResult
from textual.widgets import Static

ease = EASING[DEFAULT_EASING]


class AnimationApp(App):
    def compose(self) -> ComposeResult:
        self.box = Static("Hello, World!")
        self.box.styles.background = "red"
        self.box.styles.color = "black"
        self.box.styles.padding = (1, 2)
        yield self.box

    def key_1(self):
        self.box.styles.opacity = 1 - ease(0.25)

    def key_2(self):
        self.box.styles.opacity = 1 - ease(0.5)

    def key_3(self):
        self.box.styles.opacity = 1 - ease(0.75)


if __name__ == "__main__":
    app = AnimationApp()
    app.run()



import pytest
from rich.console import Console
from rich.segment import Segment
from rich.style import Style
from rich.text import Text

from textual._border import render_border_label, render_row
from textual.widget import Widget

_EMPTY_STYLE = Style()
_BLANK_SEGMENT = Segment(" ", _EMPTY_STYLE)
_WIDE_CONSOLE = Console(width=9999)


def test_border_render_row():
    style = Style.parse("red")
    row = (Segment("┏", style), Segment("━", style), Segment("┓", style))

    assert list(render_row(row, 5, False, False, ())) == [
        Segment(row[1].text * 5, row[1].style)
    ]
    assert list(render_row(row, 5, True, False, ())) == [
        row[0],
        Segment(row[1].text * 4, row[1].style),
    ]
    assert list(render_row(row, 5, False, True, ())) == [
        Segment(row[1].text * 4, row[1].style),
        row[2],
    ]
    assert list(render_row(row, 5, True, True, ())) == [
        row[0],
        Segment(row[1].text * 3, row[1].style),
        row[2],
    ]


def test_border_title_single_line():
    """The border_title gets set to a single line even when multiple lines are provided."""
    widget = Widget()

    assert widget.border_title is None

    widget.border_title = None
    assert widget.border_title is None

    widget.border_title = ""
    assert widget.border_title == ""

    widget.border_title = "How is life\ngoing for you?"
    assert widget.border_title == "How is life"

    widget.border_title = "How is life\n\rgoing for you?"
    assert widget.border_title == "How is life"

    widget.border_title = "Sorry you \r\n have to \n read this."
    assert widget.border_title == "Sorry you "

    widget.border_title = "[red]This also \n works with markup \n involved.[/]"
    assert widget.border_title == "[red]This also [/red]"

    widget.border_title = Text.from_markup("[bold]Hello World")
    assert widget.border_title == "[bold]Hello World[/bold]"


def test_border_subtitle_single_line():
    """The border_subtitle gets set to a single line even when multiple lines are provided."""
    widget = Widget()

    widget.border_subtitle = ""
    assert widget.border_subtitle == ""

    widget.border_subtitle = "How is life\ngoing for you?"
    assert widget.border_subtitle == "How is life"

    widget.border_subtitle = "How is life\n\rgoing for you?"
    assert widget.border_subtitle == "How is life"

    widget.border_subtitle = "Sorry you \r\n have to \n read this."
    assert widget.border_subtitle == "Sorry you "

    widget.border_subtitle = "[red]This also \n works with markup \n involved.[/]"
    assert widget.border_subtitle == "[red]This also [/red]"

    widget.border_subtitle = Text.from_markup("[bold]Hello World")
    assert widget.border_subtitle == "[bold]Hello World[/bold]"


@pytest.mark.parametrize(
    ["width", "has_left_corner", "has_right_corner"],
    [
        (10, True, True),
        (10, True, False),
        (10, False, False),
        (10, False, True),
        (1, True, True),
        (1, True, False),
        (1, False, False),
        (1, False, True),
    ],
)
def test_render_border_label_empty_label_skipped(
    width: int, has_left_corner: bool, has_right_corner: bool
):
    """Test that we get an empty list of segments if there is no label to display."""

    assert [] == list(
        render_border_label(
            (Text(""), Style()),
            True,
            "round",
            width,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _WIDE_CONSOLE,
            has_left_corner,
            has_right_corner,
        )
    )


@pytest.mark.parametrize(
    ["label", "width", "has_left_corner", "has_right_corner"],
    [
        ("hey", 2, True, True),
        ("hey", 2, True, False),
        ("hey", 2, False, True),
        ("hey", 3, True, True),
        ("hey", 4, True, True),
    ],
)
def test_render_border_label_skipped_if_narrow(
    label: str, width: int, has_left_corner: bool, has_right_corner: bool
):
    """Test that we skip rendering a label when we do not have space for it.

    In order for us to have enough space for the label, we need to have space for the
    corners that we need (none, just one, or both) and we need to be able to have two
    blank spaces around the label (one on each side).
    If we don't have space for all of these, we skip the label altogether.
    """

    assert [] == list(
        render_border_label(
            (Text.from_markup(label), Style()),
            True,
            "round",
            width,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _WIDE_CONSOLE,
            has_left_corner,
            has_right_corner,
        )
    )


@pytest.mark.parametrize(
    "label",
    [
        "Why did the scarecrow",
        "win a Nobel prize?",
        "because it was outstanding",
        "in its field.",
    ],
)
def test_render_border_label_wide_plain(label: str):
    """Test label rendering in a wide area with no styling."""

    BIG_NUM = 9999
    args = (
        True,
        "round",
        BIG_NUM,
        _EMPTY_STYLE,
        _EMPTY_STYLE,
        _EMPTY_STYLE,
        _WIDE_CONSOLE,
        True,
        True,
    )
    segments = render_border_label((Text.from_markup(label), Style()), *args)
    (segment,) = segments

    assert segment == Segment(f" {label} ", None)


@pytest.mark.parametrize(
    "label",
    [
        "[b][/]",
        "[i b][/]",
        "[white on red][/]",
        "[blue]",
    ],
)
def test_render_border_empty_text_with_markup(label: str):
    """Test label rendering if there is no text but some markup."""
    assert [] == list(
        render_border_label(
            (Text.from_markup(label), Style()),
            True,
            "round",
            999,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _EMPTY_STYLE,
            _WIDE_CONSOLE,
            True,
            True,
        )
    )


def test_render_border_label():
    """Test label rendering with styling, with and without overflow."""

    label = "[b][on red]What [i]is up[/on red] with you?[/]"
    border_style = Style.parse("green on blue")

    # Implicit test on the number of segments returned:
    blank1, what, is_up, with_you, blank2 = render_border_label(
        (Text.from_markup(label), Style()),
        True,
        "round",
        9999,
        _EMPTY_STYLE,
        _EMPTY_STYLE,
        border_style,
        _WIDE_CONSOLE,
        True,
        True,
    )

    expected_blank = Segment(" ", border_style)
    assert blank1 == expected_blank
    assert blank2 == expected_blank

    what_style = Style.parse("b on red")
    expected_what = Segment("What ", border_style + what_style)
    assert what == expected_what

    is_up_style = Style.parse("b on red i")
    expected_is_up = Segment("is up", border_style + is_up_style)
    assert is_up == expected_is_up

    with_you_style = Style.parse("b i")
    expected_with_you = Segment(" with you?", border_style + with_you_style)
    assert with_you == expected_with_you

    blank1, what, blank2 = render_border_label(
        (Text.from_markup(label), Style()),
        True,
        "round",
        5 + 4,  # 5 where "What…" fits + 2 for the blank spaces + 2 for the corners.
        _EMPTY_STYLE,
        _EMPTY_STYLE,
        border_style,
        _WIDE_CONSOLE,
        True,  # This corner costs 2 cells.
        True,  # This corner costs 2 cells.
    )

    assert blank1 == expected_blank
    assert blank2 == expected_blank

    expected_what = Segment("What…", border_style + what_style)
    assert what == expected_what



import pytest

from textual._two_way_dict import TwoWayDict


@pytest.fixture
def two_way_dict():
    return TwoWayDict(
        {
            1: 10,
            2: 20,
            3: 30,
        }
    )


def test_get(two_way_dict):
    assert two_way_dict.get(1) == 10


def test_get_key(two_way_dict):
    assert two_way_dict.get_key(30) == 3


def test_set_item(two_way_dict):
    two_way_dict[40] = 400
    assert two_way_dict.get(40) == 400
    assert two_way_dict.get_key(400) == 40


def test_len(two_way_dict):
    assert len(two_way_dict) == 3


def test_delitem(two_way_dict):
    assert two_way_dict.get(3) == 30
    assert two_way_dict.get_key(30) == 3
    del two_way_dict[3]
    assert two_way_dict.get(3) is None
    assert two_way_dict.get_key(30) is None


def test_contains(two_way_dict):
    assert 1 in two_way_dict
    assert 10 not in two_way_dict



import inspect

from textual.widgets import TextArea


def test_code_editor_parameters_kept_up_to_date():
    """Meta test to ensure the `TextArea.code_editor` convenience constructor
    is kept up to date with changes to the `TextArea.__init__` parameters.
    """
    text_area_params = inspect.signature(TextArea.__init__).parameters
    code_editor_params = inspect.signature(TextArea.code_editor).parameters
    expected_diffs = ["theme", "soft_wrap", "tab_behavior", "show_line_numbers"]
    for param in text_area_params:
        if param == "self":
            continue
        assert param in code_editor_params
        if param not in expected_diffs:
            assert code_editor_params[param] == text_area_params[param]



from textual.app import App, ComposeResult
from textual.message import Message
from textual.widget import Widget


async def test_message_inheritance_namespace():
    """Inherited messages get their correct namespaces.

    Regression test for https://github.com/Textualize/textual/issues/1814.
    """

    class BaseWidget(Widget):
        class Fired(Message):
            pass

        def trigger(self) -> None:
            self.post_message(self.Fired())

    class Left(BaseWidget):
        class Fired(BaseWidget.Fired):
            pass

    class Right(BaseWidget):
        class Fired(BaseWidget.Fired):
            pass

    class DummyWidget(Widget):
        # ensure that referencing a message type in other class scopes
        # doesn't break the namespace
        _event = Left.Fired

    handlers_called = []

    class MessageInheritanceApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Left()
            yield Right()

        def on_left_fired(self):
            handlers_called.append("left")

        def on_right_fired(self):
            handlers_called.append("right")

    app = MessageInheritanceApp()
    async with app.run_test():
        app.query_one(Left).trigger()
        app.query_one(Right).trigger()

    assert handlers_called == ["left", "right"]



from textual.app import App, ComposeResult
from textual.widgets import Static


class LayersExample(App):
    CSS_PATH = "layers.tcss"

    def compose(self) -> ComposeResult:
        yield Static("box1 (layer = above)", id="box1")
        yield Static("box2 (layer = below)", id="box2")


if __name__ == "__main__":
    app = LayersExample()
    app.run()



from __future__ import annotations

import pytest

from textual.css.tokenize import tokenize
from textual.css.tokenizer import Token, TokenError

VALID_VARIABLE_NAMES = [
    "warning-text",
    "warning_text",
    "warningtext1",
    "1warningtext",
    "WarningText1",
    "warningtext_",
    "warningtext-",
    "_warningtext",
    "-warningtext",
]


@pytest.mark.parametrize("name", VALID_VARIABLE_NAMES)
def test_variable_declaration_valid_names(name):
    css = f"${name}: black on red;"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value=f"${name}:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 14),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="black",
            read_from=("", ""),
            code=css,
            location=(0, 15),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 20),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="on",
            read_from=("", ""),
            code=css,
            location=(0, 21),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 23),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="red",
            read_from=("", ""),
            code=css,
            location=(0, 24),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 27),
            referenced_by=None,
        ),
    ]


def test_variable_declaration_multiple_values():
    css = "$x: 2vw\t4% 6s  red;"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="scalar",
            value="2vw",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value="\t",
            read_from=("", ""),
            code=css,
            location=(0, 7),
            referenced_by=None,
        ),
        Token(
            name="scalar",
            value="4%",
            read_from=("", ""),
            code=css,
            location=(0, 8),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 10),
            referenced_by=None,
        ),
        Token(
            name="duration",
            value="6s",
            read_from=("", ""),
            code=css,
            location=(0, 11),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value="  ",
            read_from=("", ""),
            code=css,
            location=(0, 13),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="red",
            read_from=("", ""),
            code=css,
            location=(0, 15),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 18),
            referenced_by=None,
        ),
    ]


def test_single_line_comment():
    css = """\
# Ignored
#foo { # Ignored
    color: red; # Also ignored
} # Nada"""
    # Check the css parses
    # list(parse(css, "<foo>"))
    result = list(tokenize(css, ("", "")))

    print(result)
    expected = [
        Token(
            name="whitespace",
            value="\n",
            read_from=("", ""),
            code=css,
            location=(0, 9),
        ),
        Token(
            name="selector_start_id",
            value="#foo",
            read_from=("", ""),
            code=css,
            location=(1, 0),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(1, 4),
        ),
        Token(
            name="declaration_set_start",
            value="{",
            read_from=("", ""),
            code=css,
            location=(1, 5),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(1, 6),
        ),
        Token(
            name="whitespace",
            value="\n",
            read_from=("", ""),
            code=css,
            location=(1, 16),
        ),
        Token(
            name="whitespace",
            value="    ",
            read_from=("", ""),
            code=css,
            location=(2, 0),
        ),
        Token(
            name="declaration_name",
            value="color:",
            read_from=("", ""),
            code=css,
            location=(2, 4),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(2, 10),
        ),
        Token(
            name="token",
            value="red",
            read_from=("", ""),
            code=css,
            location=(2, 11),
        ),
        Token(
            name="declaration_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(2, 14),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(2, 15),
        ),
        Token(
            name="whitespace",
            value="\n",
            read_from=("", ""),
            code=css,
            location=(2, 30),
        ),
        Token(
            name="declaration_set_end",
            value="}",
            read_from=("", ""),
            code=css,
            location=(3, 0),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(3, 1),
        ),
    ]
    assert result == expected


def test_variable_declaration_comment_ignored():
    css = "$x: red; /* comment */"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="red",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 7),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 8),
            referenced_by=None,
        ),
    ]


def test_variable_declaration_comment_interspersed_ignored():
    css = "$x: re/* comment */d;"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="re",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="d",
            read_from=("", ""),
            code=css,
            location=(0, 19),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 20),
            referenced_by=None,
        ),
    ]


def test_variable_declaration_no_semicolon():
    css = "$x: 1\n$y: 2"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="number",
            value="1",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value="\n",
            read_from=("", ""),
            code=css,
            location=(0, 5),
            referenced_by=None,
        ),
        Token(
            name="variable_name",
            value="$y:",
            read_from=("", ""),
            code=css,
            location=(1, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(1, 3),
            referenced_by=None,
        ),
        Token(
            name="number",
            value="2",
            read_from=("", ""),
            code=css,
            location=(1, 4),
            referenced_by=None,
        ),
    ]


def test_variable_declaration_invalid_value():
    css = "$x:(@$12x)"
    with pytest.raises(TokenError):
        list(tokenize(css, ("", "")))


def test_variables_declarations_amongst_rulesets():
    css = "$x:1; .thing{text:red;} $y:2;"
    tokens = list(tokenize(css, ("", "")))
    assert tokens == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="number",
            value="1",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 5),
            referenced_by=None,
        ),
        Token(
            name="selector_start_class",
            value=".thing",
            read_from=("", ""),
            code=css,
            location=(0, 6),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_start",
            value="{",
            read_from=("", ""),
            code=css,
            location=(0, 12),
            referenced_by=None,
        ),
        Token(
            name="declaration_name",
            value="text:",
            read_from=("", ""),
            code=css,
            location=(0, 13),
            referenced_by=None,
        ),
        Token(
            name="token",
            value="red",
            read_from=("", ""),
            code=css,
            location=(0, 18),
            referenced_by=None,
        ),
        Token(
            name="declaration_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 21),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_end",
            value="}",
            read_from=("", ""),
            code=css,
            location=(0, 22),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 23),
            referenced_by=None,
        ),
        Token(
            name="variable_name",
            value="$y:",
            read_from=("", ""),
            code=css,
            location=(0, 24),
            referenced_by=None,
        ),
        Token(
            name="number",
            value="2",
            read_from=("", ""),
            code=css,
            location=(0, 27),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 28),
            referenced_by=None,
        ),
    ]


def test_variables_reference_in_rule_declaration_value():
    css = ".warn{text: $warning;}"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="selector_start_class",
            value=".warn",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_start",
            value="{",
            read_from=("", ""),
            code=css,
            location=(0, 5),
            referenced_by=None,
        ),
        Token(
            name="declaration_name",
            value="text:",
            read_from=("", ""),
            code=css,
            location=(0, 6),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 11),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$warning",
            read_from=("", ""),
            code=css,
            location=(0, 12),
            referenced_by=None,
        ),
        Token(
            name="declaration_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 20),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_end",
            value="}",
            read_from=("", ""),
            code=css,
            location=(0, 21),
            referenced_by=None,
        ),
    ]


def test_variables_reference_in_rule_declaration_value_multiple():
    css = ".card{padding: $pad-y $pad-x;}"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="selector_start_class",
            value=".card",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_start",
            value="{",
            read_from=("", ""),
            code=css,
            location=(0, 5),
            referenced_by=None,
        ),
        Token(
            name="declaration_name",
            value="padding:",
            read_from=("", ""),
            code=css,
            location=(0, 6),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 14),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$pad-y",
            read_from=("", ""),
            code=css,
            location=(0, 15),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 21),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$pad-x",
            read_from=("", ""),
            code=css,
            location=(0, 22),
            referenced_by=None,
        ),
        Token(
            name="declaration_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 28),
            referenced_by=None,
        ),
        Token(
            name="declaration_set_end",
            value="}",
            read_from=("", ""),
            code=css,
            location=(0, 29),
            referenced_by=None,
        ),
    ]


def test_variables_reference_in_variable_declaration():
    css = "$x: $y;"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$y",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value=";",
            read_from=("", ""),
            code=css,
            location=(0, 6),
            referenced_by=None,
        ),
    ]


def test_variable_references_in_variable_declaration_multiple():
    css = "$x: $y  $z\n"
    assert list(tokenize(css, ("", ""))) == [
        Token(
            name="variable_name",
            value="$x:",
            read_from=("", ""),
            code=css,
            location=(0, 0),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=css,
            location=(0, 3),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$y",
            read_from=("", ""),
            code=css,
            location=(0, 4),
            referenced_by=None,
        ),
        Token(
            name="whitespace",
            value="  ",
            read_from=("", ""),
            code=css,
            location=(0, 6),
            referenced_by=None,
        ),
        Token(
            name="variable_ref",
            value="$z",
            read_from=("", ""),
            code=css,
            location=(0, 8),
            referenced_by=None,
        ),
        Token(
            name="variable_value_end",
            value="\n",
            read_from=("", ""),
            code=css,
            location=(0, 10),
            referenced_by=None,
        ),
    ]


def test_allow_new_lines():
    css = ".foo{margin: 1\n1 0 0}"
    tokens = list(tokenize(css, ("", "")))
    print(repr(tokens))
    expected = [
        Token(
            name="selector_start_class",
            value=".foo",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 0),
        ),
        Token(
            name="declaration_set_start",
            value="{",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 4),
        ),
        Token(
            name="declaration_name",
            value="margin:",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 5),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 12),
        ),
        Token(
            name="number",
            value="1",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 13),
        ),
        Token(
            name="whitespace",
            value="\n",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(0, 14),
        ),
        Token(
            name="number",
            value="1",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 0),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 1),
        ),
        Token(
            name="number",
            value="0",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 2),
        ),
        Token(
            name="whitespace",
            value=" ",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 3),
        ),
        Token(
            name="number",
            value="0",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 4),
        ),
        Token(
            name="declaration_set_end",
            value="}",
            read_from=("", ""),
            code=".foo{margin: 1\n1 0 0}",
            location=(1, 5),
        ),
    ]
    assert list(tokenize(css, ("", ""))) == expected


@pytest.mark.parametrize(
    ["pseudo_class", "expected"],
    [
        ("blue", "blur"),
        ("br", "blur"),
        ("canfocus", "can-focus"),
        ("can_focus", "can-focus"),
        ("can-foc", "can-focus"),
        ("drk", "dark"),
        ("ark", "dark"),
        ("disssabled", "disabled"),
        ("enalbed", "enabled"),
        ("focoswithin", "focus-within"),
        ("focus_whitin", "focus-within"),
        ("fcus", "focus"),
        ("huver", "hover"),
        ("LIght", "light"),
    ],
)
def test_did_you_mean_pseudo_classes(pseudo_class: str, expected: str) -> None:
    """Make sure we get the correct suggestion for pseudo-classes with typos."""

    css = f"""
    Button:{pseudo_class} {{
        background: red;
    }}
    """

    with pytest.raises(TokenError) as err:
        list(tokenize(css, ("", "")))

    assert f"unknown pseudo-class {pseudo_class!r}" in str(err.value)
    assert f"did you mean {expected!r}" in str(err.value)


@pytest.mark.parametrize(
    ["pseudo_class", "expected"],
    [
        ("blue", "blur"),
        ("br", "blur"),
        ("canfocus", "can-focus"),
        ("can_focus", "can-focus"),
        ("can-foc", "can-focus"),
        ("drk", "dark"),
        ("ark", "dark"),
        ("disssabled", "disabled"),
        ("enalbed", "enabled"),
        ("focoswithin", "focus-within"),
        ("focus_whitin", "focus-within"),
        ("fcus", "focus"),
        ("huver", "hover"),
        ("LIght", "light"),
    ],
)
def test_did_you_mean_pseudo_classes_in_nested_css(
    pseudo_class: str, expected: str
) -> None:
    """Test that we get nice errors for pseudo-classes with typos in nested TCSS.

    When implementing pseudo-class support in nested TCSS
    (https://github.com/Textualize/textual/issues/4039), the first iterations didn't
    preserve this so we add these tests to make sure we don't take this feature away
    unintentionally.
    """

    css = f"""
    Screen {{
        Button:{pseudo_class} {{
            background: red;
        }}
    }}
    """

    with pytest.raises(TokenError) as err:
        list(tokenize(css, ("", "")))

    assert f"unknown pseudo-class {pseudo_class!r}" in str(err.value)
    assert f"did you mean {expected!r}" in str(err.value)



import pytest

from textual.widgets import Placeholder
from textual.widgets._placeholder import InvalidPlaceholderVariant


def test_invalid_placeholder_variant():
    with pytest.raises(InvalidPlaceholderVariant):
        Placeholder(variant="this is clearly not a valid variant!")


def test_invalid_reactive_variant_change():
    p = Placeholder()
    with pytest.raises(InvalidPlaceholderVariant):
        p.variant = "this is clearly not a valid variant!"



from textual.app import App, ComposeResult
from textual.color import Color
from textual.widgets import Footer, Static


class Bar(Static):
    pass


class BindingApp(App):
    CSS_PATH = "binding01.tcss"
    BINDINGS = [
        ("r", "add_bar('red')", "Add Red"),
        ("g", "add_bar('green')", "Add Green"),
        ("b", "add_bar('blue')", "Add Blue"),
    ]

    def compose(self) -> ComposeResult:
        yield Footer()

    def action_add_bar(self, color: str) -> None:
        bar = Bar(color)
        bar.styles.background = Color.parse(color).with_alpha(0.5)
        self.mount(bar)
        self.call_after_refresh(self.screen.scroll_end, animate=False)


if __name__ == "__main__":
    app = BindingApp()
    app.run()



from asyncio import sleep
from random import randint

from textual import work
from textual.app import App, ComposeResult
from textual.widgets import DataTable

ROWS = [
    ("lane", "swimmer", "country", "time"),
    (4, "Joseph Schooling", "Singapore", 50.39),
    (2, "Michael Phelps", "United States", 51.14),
    (5, "Chad le Clos", "South Africa", 51.14),
    (6, "László Cseh", "Hungary", 51.14),
    (3, "Li Zhuhao", "China", 51.26),
    (8, "Mehdy Metella", "France", 51.58),
    (7, "Tom Shields", "United States", 51.73),
    (1, "Aleksandr Sadovnikov", "Russia", 51.84),
    (10, "Darren Burns", "Scotland", 51.84),
]


class DataApp(App):
    CSS = """
    Screen {
        layout: grid;
        grid-size: 2;
    }
    DataTable {
        height: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield DataTable()
        yield DataTable()
        yield DataTable()
        yield DataTable()

    def on_mount(self) -> None:
        for data_table in self.query(DataTable):
            data_table.loading = True  # (1)!
            self.load_data(data_table)

    @work
    async def load_data(self, data_table: DataTable) -> None:
        await sleep(randint(2, 10))  # (2)!
        data_table.add_columns(*ROWS[0])
        data_table.add_rows(ROWS[1:])
        data_table.loading = False  # (3)!


if __name__ == "__main__":
    app = DataApp()
    app.run()



from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Header, Select

LINES = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.""".splitlines()


class SelectApp(App):
    CSS_PATH = "select.tcss"

    def compose(self) -> ComposeResult:
        yield Header()
        yield Select.from_values(LINES)

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed) -> None:
        self.title = str(event.value)


if __name__ == "__main__":
    app = SelectApp()
    app.run()



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Flag, auto
from typing import TYPE_CHECKING, Iterable, Pattern

from rich.console import Console, ConsoleOptions, RenderableType
from rich.console import RenderResult as RichRenderResult
from rich.segment import Segment
from rich.text import Text
from typing_extensions import Literal

from textual import events
from textual._segment_tools import line_crop

if TYPE_CHECKING:
    from textual.app import RenderResult

from textual.reactive import var
from textual.validation import ValidationResult, Validator
from textual.widgets._input import Input

InputValidationOn = Literal["blur", "changed", "submitted"]
"""Possible messages that trigger input validation."""


class _CharFlags(Flag):
    """Misc flags for a single template character definition"""

    NONE = 0
    """Empty flags value"""

    REQUIRED = auto()
    """Is this character required for validation?"""

    SEPARATOR = auto()
    """Is this character a separator?"""

    UPPERCASE = auto()
    """Char is forced to be uppercase"""

    LOWERCASE = auto()
    """Char is forced to be lowercase"""


_TEMPLATE_CHARACTERS = {
    "A": (r"[A-Za-z]", _CharFlags.REQUIRED),
    "a": (r"[A-Za-z]", None),
    "N": (r"[A-Za-z0-9]", _CharFlags.REQUIRED),
    "n": (r"[A-Za-z0-9]", None),
    "X": (r"[^ ]", _CharFlags.REQUIRED),
    "x": (r"[^ ]", None),
    "9": (r"[0-9]", _CharFlags.REQUIRED),
    "0": (r"[0-9]", None),
    "D": (r"[1-9]", _CharFlags.REQUIRED),
    "d": (r"[1-9]", None),
    "#": (r"[0-9+\-]", None),
    "H": (r"[A-Fa-f0-9]", _CharFlags.REQUIRED),
    "h": (r"[A-Fa-f0-9]", None),
    "B": (r"[0-1]", _CharFlags.REQUIRED),
    "b": (r"[0-1]", None),
}


class _InputRenderable:
    """Render the input content."""

    def __init__(self, input: Input, cursor_visible: bool) -> None:
        self.input = input
        self.cursor_visible = cursor_visible

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> RichRenderResult:
        input = self.input
        result = input._value
        width = input.content_size.width

        # Add the completion with a faded style.
        value = input.value
        value_length = len(value)
        template = input._template
        style = input.get_component_rich_style("input--placeholder")
        result += Text(
            template.mask[value_length:],
            style,
        )
        for index, (char, char_definition) in enumerate(zip(value, template.template)):
            if char == " ":
                result.stylize(style, index, index + 1)

        if self.cursor_visible and input.has_focus:
            if input._cursor_at_end:
                result.pad_right(1)
            cursor_style = input.get_component_rich_style("input--cursor")
            cursor = input.cursor_position
            result.stylize(cursor_style, cursor, cursor + 1)

        segments = list(result.render(console))
        line_length = Segment.get_line_length(segments)
        if line_length < width:
            segments = Segment.adjust_line_length(segments, width)
            line_length = width

        line = line_crop(
            list(segments),
            input.view_position,
            input.view_position + width,
            line_length,
        )
        yield from line


class _Template(Validator):
    """Template mask enforcer."""

    @dataclass
    class CharDefinition:
        """Holds data for a single char of the template mask."""

        pattern: Pattern[str]
        """Compiled regular expression to check for matches."""

        flags: _CharFlags = _CharFlags.NONE
        """Flags defining special behaviors"""

        char: str = ""
        """Mask character (separator or blank or placeholder)"""

    def __init__(self, input: Input, template_str: str) -> None:
        """Initialise the mask enforcer, which is also a subclass of `Validator`.

        Args:
            input: The `MaskedInput` that owns this object.
            template_str: Template string controlling masked input behavior.
        """
        self.input = input
        self.template: list[_Template.CharDefinition] = []
        self.blank: str = " "
        escaped = False
        flags = _CharFlags.NONE
        template_chars: list[str] = list(template_str)

        while template_chars:
            char = template_chars.pop(0)
            if escaped:
                char_definition = self.CharDefinition(
                    re.compile(re.escape(char)), _CharFlags.SEPARATOR, char
                )
                escaped = False
            else:
                if char == "\\":
                    escaped = True
                    continue
                elif char == ";":
                    break

                new_flags = {
                    ">": _CharFlags.UPPERCASE,
                    "<": _CharFlags.LOWERCASE,
                    "!": _CharFlags.NONE,
                }.get(char, None)
                if new_flags is not None:
                    flags = new_flags
                    continue

                pattern, required_flag = _TEMPLATE_CHARACTERS.get(char, (None, None))
                if pattern:
                    char_flags = (
                        _CharFlags.REQUIRED if required_flag else _CharFlags.NONE
                    )
                    char_definition = self.CharDefinition(
                        re.compile(pattern), char_flags
                    )
                else:
                    char_definition = self.CharDefinition(
                        re.compile(re.escape(char)), _CharFlags.SEPARATOR, char
                    )

            char_definition.flags |= flags
            self.template.append(char_definition)

        if template_chars:
            self.blank = template_chars[0]

        if all(
            (_CharFlags.SEPARATOR in char_definition.flags)
            for char_definition in self.template
        ):
            raise ValueError(
                "Template must contain at least one non-separator character"
            )

        self.update_mask(input.placeholder)

    def validate(self, value: str) -> ValidationResult:
        """Checks if `value` matches this template, always returning a ValidationResult.

        Args:
            value: The string value to be validated.

        Returns:
            A ValidationResult with the validation outcome.

        """
        if self.check(value.ljust(len(self.template), chr(0)), False):
            return self.success()
        else:
            return self.failure("Value does not match template!", value)

    def check(self, value: str, allow_space: bool) -> bool:
        """Checks if `value matches this template, but returns result as a bool.

        Args:
            value: The string value to be validated.
            allow_space: Consider space character in `value` as valid.

        Returns:
            True if `value` is valid for this template, False otherwise.
        """
        for char, char_definition in zip(value, self.template):
            if (
                (_CharFlags.REQUIRED in char_definition.flags)
                and (not char_definition.pattern.match(char))
                and ((char != " ") or not allow_space)
            ):
                return False
        return True

    def insert_separators(self, value: str, cursor_position: int) -> tuple[str, int]:
        """Automatically inserts separators in `value` at `cursor_position` if expected, eventually advancing
        the current cursor position.

        Args:
            value: Current control value entered by user.
            cursor_position: Where to start inserting separators (if any).

        Returns:
            A tuple in the form `(value, cursor_position)` with new value and possibly advanced cursor position.
        """
        while cursor_position < len(self.template) and (
            _CharFlags.SEPARATOR in self.template[cursor_position].flags
        ):
            value = (
                value[:cursor_position]
                + self.template[cursor_position].char
                + value[cursor_position + 1 :]
            )
            cursor_position += 1
        return value, cursor_position

    def insert_text_at_cursor(self, text: str) -> str | None:
        """Inserts `text` at current cursor position. If not present in `text`, any expected separator is automatically
        inserted at the correct position.

        Args:
            text: The text to be inserted.

        Returns:
            A tuple in the form `(value, cursor_position)` with the new control value and current cursor position if
                `text` matches the template, None otherwise.
        """
        value = self.input.value
        cursor_position = self.input.cursor_position
        separators = set(
            [
                char_definition.char
                for char_definition in self.template
                if _CharFlags.SEPARATOR in char_definition.flags
            ]
        )
        for char in text:
            if char in separators:
                if char == self.next_separator(cursor_position):
                    prev_position = self.prev_separator_position(cursor_position)
                    if (cursor_position > 0) and (prev_position != cursor_position - 1):
                        next_position = self.next_separator_position(cursor_position)
                        while cursor_position < next_position + 1:
                            if (
                                _CharFlags.SEPARATOR
                                in self.template[cursor_position].flags
                            ):
                                char = self.template[cursor_position].char
                            else:
                                char = " "
                            value = (
                                value[:cursor_position]
                                + char
                                + value[cursor_position + 1 :]
                            )
                            cursor_position += 1
                continue
            if cursor_position >= len(self.template):
                break
            char_definition = self.template[cursor_position]
            assert _CharFlags.SEPARATOR not in char_definition.flags
            if not char_definition.pattern.match(char):
                return None
            if _CharFlags.LOWERCASE in char_definition.flags:
                char = char.lower()
            elif _CharFlags.UPPERCASE in char_definition.flags:
                char = char.upper()
            value = value[:cursor_position] + char + value[cursor_position + 1 :]
            cursor_position += 1
            value, cursor_position = self.insert_separators(value, cursor_position)
        return value, cursor_position

    def move_cursor(self, delta: int) -> None:
        """Moves the cursor position by `delta` characters, skipping separators if
        running over them.

        Args:
            delta: The number of characters to move; positive moves right, negative
                moves left.
        """
        cursor_position = self.input.cursor_position
        if delta < 0 and all(
            [
                (_CharFlags.SEPARATOR in char_definition.flags)
                for char_definition in self.template[:cursor_position]
            ]
        ):
            return
        cursor_position += delta
        while (
            (cursor_position >= 0)
            and (cursor_position < len(self.template))
            and (_CharFlags.SEPARATOR in self.template[cursor_position].flags)
        ):
            cursor_position += delta
        self.input.cursor_position = cursor_position

    def delete_at_position(self, position: int | None = None) -> None:
        """Deletes character at `position`.

        Args:
            position: Position within the control value where to delete a character;
                if None the current cursor position is used.
        """
        value = self.input.value
        if position is None:
            position = self.input.cursor_position
        cursor_position = position
        if cursor_position < len(self.template):
            assert _CharFlags.SEPARATOR not in self.template[cursor_position].flags
            if cursor_position == len(value) - 1:
                value = value[:cursor_position]
            else:
                value = value[:cursor_position] + " " + value[cursor_position + 1 :]
        pos = len(value)
        while pos > 0:
            char_definition = self.template[pos - 1]
            if (_CharFlags.SEPARATOR not in char_definition.flags) and (
                value[pos - 1] != " "
            ):
                break
            pos -= 1
        value = value[:pos]
        if cursor_position > len(value):
            cursor_position = len(value)
        value, cursor_position = self.insert_separators(value, cursor_position)
        self.input.cursor_position = cursor_position
        self.input.value = value

    def at_separator(self, position: int | None = None) -> bool:
        """Checks if character at `position` is a separator.

        Args:
            position: Position within the control value where to check;
                if None the current cursor position is used.

        Returns:
            True if character is a separator, False otherwise.
        """
        if position is None:
            position = self.input.cursor_position
        if (position >= 0) and (position < len(self.template)):
            return _CharFlags.SEPARATOR in self.template[position].flags
        else:
            return False

    def prev_separator_position(self, position: int | None = None) -> int | None:
        """Obtains the position of the previous separator character starting from
        `position` within the template string.

        Args:
            position: Starting position from which to search previous separator.
                If None, current cursor position is used.

        Returns:
            The position of the previous separator, or None if no previous
                separator is found.
        """
        if position is None:
            position = self.input.cursor_position
        for index in range(position - 1, 0, -1):
            if _CharFlags.SEPARATOR in self.template[index].flags:
                return index
        else:
            return None

    def next_separator_position(self, position: int | None = None) -> int | None:
        """Obtains the position of the next separator character starting from
        `position` within the template string.

        Args:
            position: Starting position from which to search next separator.
                If None, current cursor position is used.

        Returns:
            The position of the next separator, or None if no next
                separator is found.
        """
        if position is None:
            position = self.input.cursor_position
        for index in range(position + 1, len(self.template)):
            if _CharFlags.SEPARATOR in self.template[index].flags:
                return index
        else:
            return None

    def next_separator(self, position: int | None = None) -> str | None:
        """Obtains the next separator character starting from `position`
        within the template string.

        Args:
            position: Starting position from which to search next separator.
                If None, current cursor position is used.

        Returns:
            The next separator character, or None if no next
                separator is found.
        """
        position = self.next_separator_position(position)
        if position is None:
            return None
        else:
            return self.template[position].char

    def display(self, value: str) -> str:
        """Returns `value` ready for display, with spaces replaced by
        placeholder characters.

        Args:
            value: String value to display.

        Returns:
            New string value with spaces replaced by placeholders.
        """
        result = []
        for char, char_definition in zip(value, self.template):
            if char == " ":
                char = char_definition.char
            result.append(char)
        return "".join(result)

    def update_mask(self, placeholder: str) -> None:
        """Updates template placeholder characters from `placeholder`. If
        given string is smaller than template string, template blank character
        is used to fill remaining template placeholder characters.

        Args:
            placeholder: New placeholder string.
        """
        for index, char_definition in enumerate(self.template):
            if _CharFlags.SEPARATOR not in char_definition.flags:
                if index < len(placeholder):
                    char_definition.char = placeholder[index]
                else:
                    char_definition.char = self.blank

    @property
    def mask(self) -> str:
        """Property returning the template placeholder mask."""
        return "".join([char_definition.char for char_definition in self.template])

    @property
    def empty_mask(self) -> str:
        """Property returning the template placeholder mask with all non-separators replaced by space."""
        return "".join(
            [
                (
                    " "
                    if (_CharFlags.SEPARATOR not in char_definition.flags)
                    else char_definition.char
                )
                for char_definition in self.template
            ]
        )


class MaskedInput(Input, can_focus=True):
    """A masked text input widget."""

    template = var("")
    """Input template mask currently in use."""

    def __init__(
        self,
        template: str,
        value: str | None = None,
        placeholder: str = "",
        *,
        validators: Validator | Iterable[Validator] | None = None,
        validate_on: Iterable[InputValidationOn] | None = None,
        valid_empty: bool = False,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
    ) -> None:
        """Initialise the `Input` widget.

        Args:
            template: Template string.
            value: An optional default value for the input.
            placeholder: Optional placeholder text for the input.
            validators: An iterable of validators that the MaskedInput value will be checked against.
            validate_on: Zero or more of the values "blur", "changed", and "submitted",
                which determine when to do input validation. The default is to do
                validation for all messages.
            valid_empty: Empty values are valid.
            name: Optional name for the masked input widget.
            id: Optional ID for the widget.
            classes: Optional initial classes for the widget.
            disabled: Whether the input is disabled or not.
            tooltip: Optional tooltip.
        """
        self._template: _Template = None
        super().__init__(
            placeholder=placeholder,
            validators=validators,
            validate_on=validate_on,
            valid_empty=valid_empty,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

        self._template = _Template(self, template)
        self.template = template

        value, _ = self._template.insert_separators(value or "", 0)
        self.value = value
        if tooltip is not None:
            self.tooltip = tooltip

    def validate_value(self, value: str) -> str:
        """Validates value against template."""
        if self._template is None:
            return value
        if not self._template.check(value, True):
            raise ValueError("Value does not match template!")
        return value[: len(self._template.mask)]

    def _watch_template(self, template: str) -> None:
        """Revalidate when template changes."""
        self._template = _Template(self, template) if template else None
        if self.is_mounted:
            self._watch_value(self.value)

    def _watch_placeholder(self, placeholder: str) -> None:
        """Update template display mask when placeholder changes."""
        if self._template is not None:
            self._template.update_mask(placeholder)
            self.refresh()

    def validate(self, value: str) -> ValidationResult | None:
        """Run all the validators associated with this MaskedInput on the supplied value.

        Same as `Input.validate()` but also validates against template which acts as an
        additional implicit validator.

        Returns:
            A ValidationResult indicating whether *all* validators succeeded or not.
                That is, if *any* validator fails, the result will be an unsuccessful
                validation.
        """

        def set_classes() -> None:
            """Set classes for valid flag."""
            valid = self._valid
            self.set_class(not valid, "-invalid")
            self.set_class(valid, "-valid")

        result = super().validate(value)
        validation_results: list[ValidationResult] = [self._template.validate(value)]
        if result is not None:
            validation_results.append(result)
        combined_result = ValidationResult.merge(validation_results)
        self._valid = combined_result.is_valid
        set_classes()

        return combined_result

    def render(self) -> RenderResult:
        return _InputRenderable(self, self._cursor_visible)

    @property
    def _value(self) -> Text:
        """Value rendered as text."""
        value = self._template.display(self.value)
        return Text(value, no_wrap=True, overflow="ignore")

    async def _on_click(self, event: events.Click) -> None:
        """Ensure clicking on value does not leave cursor on a separator."""
        await super()._on_click(event)
        if self._template.at_separator():
            self._template.move_cursor(1)

    def insert_text_at_cursor(self, text: str) -> None:
        """Insert new text at the cursor, move the cursor to the end of the new text.

        Args:
            text: New text to insert.
        """

        new_value = self._template.insert_text_at_cursor(text)
        if new_value is not None:
            self.value, self.cursor_position = new_value
        else:
            self.restricted()

    def clear(self) -> None:
        """Clear the masked input."""
        self.value, self.cursor_position = self._template.insert_separators("", 0)

    def action_cursor_left(self) -> None:
        """Move the cursor one position to the left; separators are skipped."""
        self._template.move_cursor(-1)

    def action_cursor_right(self) -> None:
        """Move the cursor one position to the right; separators are skipped."""
        self._template.move_cursor(1)

    def action_home(self) -> None:
        """Move the cursor to the start of the input."""
        self._template.move_cursor(-len(self.template))

    def action_cursor_left_word(self) -> None:
        """Move the cursor left next to the previous separator. If no previous
        separator is found, moves the cursor to the start of the input."""
        if self._template.at_separator(self.cursor_position - 1):
            position = self._template.prev_separator_position(self.cursor_position - 1)
        else:
            position = self._template.prev_separator_position()
        if position:
            position += 1
        self.cursor_position = position or 0

    def action_cursor_right_word(self) -> None:
        """Move the cursor right next to the next separator. If no next
        separator is found, moves the cursor to the end of the input."""
        position = self._template.next_separator_position()
        if position is None:
            self.cursor_position = len(self._template.mask)
        else:
            self.cursor_position = position + 1

    def action_delete_right(self) -> None:
        """Delete one character at the current cursor position."""
        self._template.delete_at_position()

    def action_delete_right_word(self) -> None:
        """Delete the current character and all rightward to next separator or
        the end of the input."""
        position = self._template.next_separator_position()
        if position is not None:
            position += 1
        else:
            position = len(self.value)
        for index in range(self.cursor_position, position):
            self.cursor_position = index
            if not self._template.at_separator():
                self._template.delete_at_position()

    def action_delete_left(self) -> None:
        """Delete one character to the left of the current cursor position."""
        if self.cursor_position <= 0:
            # Cursor at the start, so nothing to delete
            return
        self._template.move_cursor(-1)
        self._template.delete_at_position()

    def action_delete_left_word(self) -> None:
        """Delete leftward of the cursor position to the previous separator or
        the start of the input."""
        if self.cursor_position <= 0:
            return
        if self._template.at_separator(self.cursor_position - 1):
            position = self._template.prev_separator_position(self.cursor_position - 1)
        else:
            position = self._template.prev_separator_position()
        if position:
            position += 1
        else:
            position = 0
        for index in range(position, self.cursor_position):
            self.cursor_position = index
            if not self._template.at_separator():
                self._template.delete_at_position()
        self.cursor_position = position

    def action_delete_left_all(self) -> None:
        """Delete all characters to the left of the cursor position."""
        if self.cursor_position > 0:
            cursor_position = self.cursor_position
            if cursor_position >= len(self.value):
                self.value = ""
            else:
                self.value = (
                    self._template.empty_mask[:cursor_position]
                    + self.value[cursor_position:]
                )
            self.cursor_position = 0



from __future__ import annotations

from typing import TYPE_CHECKING

from textual._animator import Animation, EasingFunction
from textual._types import AnimationLevel, CallbackType
from textual.css.scalar import Scalar, ScalarOffset

if TYPE_CHECKING:
    from textual.css.styles import StylesBase
    from textual.widget import Widget


class ScalarAnimation(Animation):
    def __init__(
        self,
        widget: Widget,
        styles: StylesBase,
        start_time: float,
        attribute: str,
        value: ScalarOffset | Scalar,
        duration: float | None,
        speed: float | None,
        easing: EasingFunction,
        on_complete: CallbackType | None = None,
        level: AnimationLevel = "full",
    ):
        assert (
            speed is not None or duration is not None
        ), "One of speed or duration required"
        self.widget = widget
        self.styles = styles
        self.start_time = start_time
        self.attribute = attribute
        self.final_value = value
        self.easing = easing
        self.on_complete = on_complete
        self.level = level

        size = widget.outer_size
        viewport = widget.app.size

        self.start = getattr(styles, attribute).resolve(size, viewport)
        self.destination = value.resolve(size, viewport)

        if speed is not None:
            distance = self.start.get_distance_to(self.destination)
            self.duration = distance / speed
        else:
            assert duration is not None, "Duration expected to be non-None"
            self.duration = duration

    def __call__(
        self, time: float, app_animation_level: AnimationLevel = "full"
    ) -> bool:
        factor = min(1.0, (time - self.start_time) / self.duration)
        eased_factor = self.easing(factor)

        if (
            eased_factor >= 1
            or app_animation_level == "none"
            or app_animation_level == "basic"
            and self.level == "full"
        ):
            setattr(self.styles, self.attribute, self.final_value)
            return True

        if hasattr(self.start, "blend"):
            value = self.start.blend(self.destination, eased_factor)
        else:
            value = self.start + (self.destination - self.start) * eased_factor
        current = self.styles.get_rule(self.attribute)
        if current != value:
            setattr(self.styles, self.attribute, value)

        return False

    async def stop(self, complete: bool = True) -> None:
        """Stop the animation.

        Args:
            complete: Flag to say if the animation should be taken to completion.

        Note:
            [`on_complete`][Animation.on_complete] will be called regardless
            of the value provided for `complete`.
        """
        if complete:
            setattr(self.styles, self.attribute, self.final_value)
        await self.invoke_callback()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ScalarAnimation):
            return (
                self.final_value == other.final_value
                and self.duration == other.duration
            )
        return False



from textual.app import App, ComposeResult
from textual.containers import Horizontal
from textual.widgets import Collapsible, Label


class CollapsibleApp(App[None]):
    def compose(self) -> ComposeResult:
        with Horizontal():
            with Collapsible(
                collapsed_symbol=">>>",
                expanded_symbol="v",
            ):
                yield Label("Hello, world.")

            with Collapsible(
                collapsed_symbol=">>>",
                expanded_symbol="v",
                collapsed=False,
            ):
                yield Label("Hello, world.")


if __name__ == "__main__":
    app = CollapsibleApp()
    app.run()



import pytest

from textual.app import App, ComposeResult
from textual.widgets import TextArea
from textual.widgets.text_area import Selection

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""


class TextAreaApp(App):
    def compose(self) -> ComposeResult:
        text_area = TextArea()
        text_area.load_text(TEXT)
        yield text_area


async def test_default_selection():
    """The cursor starts at (0, 0) in the document."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        assert text_area.selection == Selection.cursor((0, 0))


async def test_cursor_location_get():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((1, 1), (2, 2))
        assert text_area.cursor_location == (2, 2)


async def test_cursor_location_set():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        target = (1, 2)
        text_area.cursor_location = target
        assert text_area.selection == Selection.cursor(target)


async def test_cursor_location_set_while_selecting():
    """If you set the cursor_location while a selection is in progress,
    the start/anchor point of the selection will remain where it is."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((0, 0), (0, 2))
        target = (1, 2)
        text_area.cursor_location = target
        assert text_area.selection == Selection((0, 0), target)


async def test_move_cursor_select():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((1, 1), (2, 2))
        text_area.move_cursor((2, 3), select=True)
        assert text_area.selection == Selection((1, 1), (2, 3))


async def test_move_cursor_relative():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        text_area.move_cursor_relative(rows=1, columns=2)
        assert text_area.selection == Selection.cursor((1, 2))

        text_area.move_cursor_relative(rows=-1, columns=-2)
        assert text_area.selection == Selection.cursor((0, 0))

        text_area.move_cursor_relative(rows=1000, columns=1000)
        assert text_area.selection == Selection.cursor((4, 0))


async def test_selected_text_forward():
    """Selecting text from top to bottom results in the correct selected_text."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((0, 0), (2, 0))
        assert (
            text_area.selected_text
            == """\
I must not fear.
Fear is the mind-killer.
"""
        )


async def test_selected_text_backward():
    """Selecting text from bottom to top results in the correct selected_text."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((2, 0), (0, 0))
        assert (
            text_area.selected_text
            == """\
I must not fear.
Fear is the mind-killer.
"""
        )


async def test_selected_text_multibyte():
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("こんにちは")
        text_area.selection = Selection((0, 1), (0, 3))
        assert text_area.selected_text == "んに"


async def test_selection_clamp():
    """When you set the selection reactive, it's clamped to within the document bounds."""
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.selection = Selection((99, 99), (100, 100))
        assert text_area.selection == Selection(start=(4, 0), end=(4, 0))


@pytest.mark.parametrize(
    "start,end",
    [
        ((0, 0), (0, 0)),
        ((0, 4), (0, 3)),
        ((1, 0), (0, 16)),
    ],
)
async def test_get_cursor_left_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor(start)
        assert text_area.get_cursor_left_location() == end


@pytest.mark.parametrize(
    "start,end",
    [
        ((0, 0), (0, 1)),
        ((0, 16), (1, 0)),
        ((3, 20), (4, 0)),
        ((4, 0), (4, 0)),
    ],
)
async def test_get_cursor_right_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor(start)
        assert text_area.get_cursor_right_location() == end


@pytest.mark.parametrize(
    "start,end",
    [
        ((0, 4), (0, 0)),  # jump to start
        ((1, 2), (0, 2)),  # go to column above
        ((2, 56), (1, 24)),  # snap to end of row above
    ],
)
async def test_get_cursor_up_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor(start)
        # This is required otherwise the cursor will snap back to the
        # last location navigated to (0, 0)
        text_area.record_cursor_width()
        assert text_area.get_cursor_up_location() == end


@pytest.mark.parametrize(
    "start,end",
    [
        ((3, 4), (4, 0)),  # jump to end
        ((1, 2), (2, 2)),  # go to column above
        ((2, 56), (3, 20)),  # snap to end of row below
    ],
)
async def test_get_cursor_down_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.move_cursor(start)
        # This is required otherwise the cursor will snap back to the
        # last location navigated to (0, 0)
        text_area.record_cursor_width()
        assert text_area.get_cursor_down_location() == end


@pytest.mark.parametrize(
    "start,end",
    [
        ((0, 0), (0, 0)),
        ((0, 1), (0, 0)),
        ((0, 2), (0, 0)),
        ((0, 3), (0, 0)),
        ((0, 4), (0, 3)),
        ((0, 5), (0, 3)),
        ((0, 6), (0, 3)),
        ((0, 7), (0, 3)),
        ((0, 10), (0, 7)),
        ((1, 0), (0, 10)),
        ((1, 2), (1, 0)),
        ((1, 4), (1, 0)),
        ((1, 7), (1, 4)),
        ((1, 8), (1, 7)),
        ((1, 13), (1, 11)),
        ((1, 14), (1, 11)),
    ],
)
async def test_cursor_word_left_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("AB CD  EFG\n    HI\tJK  LM ")
        text_area.move_cursor(start)
        assert text_area.get_cursor_word_left_location() == end


@pytest.mark.parametrize(
    "start,end",
    [
        ((0, 0), (0, 2)),
        ((0, 1), (0, 2)),
        ((0, 2), (0, 5)),
        ((0, 3), (0, 5)),
        ((0, 4), (0, 5)),
        ((0, 5), (0, 10)),
        ((0, 6), (0, 10)),
        ((0, 7), (0, 10)),
        ((0, 10), (1, 0)),
        ((1, 0), (1, 6)),
        ((1, 2), (1, 6)),
        ((1, 4), (1, 6)),
        ((1, 7), (1, 9)),
        ((1, 8), (1, 9)),
        ((1, 13), (1, 14)),
        ((1, 14), (1, 14)),
    ],
)
async def test_cursor_word_right_location(start, end):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text("AB CD  EFG\n    HI\tJK  LM ")
        text_area.move_cursor(start)
        assert text_area.get_cursor_word_right_location() == end


@pytest.mark.parametrize(
    "content,expected_selection",
    [
        ("123\n456\n789", Selection((0, 0), (2, 3))),
        ("123\n456\n789\n", Selection((0, 0), (3, 0))),
        ("", Selection((0, 0), (0, 0))),
    ],
)
async def test_select_all(content, expected_selection):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text(content)

        text_area.select_all()

        assert text_area.selection == expected_selection


@pytest.mark.parametrize(
    "index,content,expected_selection",
    [
        (1, "123\n456\n789\n", Selection((1, 0), (1, 3))),
        (2, "123\n456\n789\n", Selection((2, 0), (2, 3))),
        (3, "123\n456\n789\n", Selection((3, 0), (3, 0))),
        (1000, "123\n456\n789\n", Selection.cursor((0, 0))),
        (0, "", Selection((0, 0), (0, 0))),
    ],
)
async def test_select_line(index, content, expected_selection):
    app = TextAreaApp()
    async with app.run_test():
        text_area = app.query_one(TextArea)
        text_area.load_text(content)

        text_area.select_line(index)

        assert text_area.selection == expected_selection


async def test_cursor_screen_offset_and_terminal_cursor_position_update():
    class TextAreaCursorScreenOffset(App):
        def compose(self) -> ComposeResult:
            yield TextArea.code_editor("abc\ndef")

    app = TextAreaCursorScreenOffset()
    async with app.run_test():
        text_area = app.query_one(TextArea)

        assert app.cursor_position == (5, 1)

        text_area.cursor_location = (1, 1)

        assert text_area.cursor_screen_offset == (6, 2)

        # Also ensure that this update has been reported back to the app
        # for the benefit of IME/emoji popups.
        assert app.cursor_position == (6, 2)


async def test_cursor_screen_offset_and_terminal_cursor_position_scrolling():
    class TextAreaCursorScreenOffset(App):
        def compose(self) -> ComposeResult:
            yield TextArea.code_editor("AB\nAB\nAB\nAB\nAB\nAB\n")

    app = TextAreaCursorScreenOffset()
    async with app.run_test(size=(80, 2)) as pilot:
        text_area = app.query_one(TextArea)

        assert app.cursor_position == (5, 1)

        text_area.cursor_location = (5, 0)
        await pilot.pause()

        assert text_area.cursor_screen_offset == (5, 1)
        assert app.cursor_position == (5, 1)



from rich.console import Console

from textual.app import App
from textual.widgets import Input


async def test_input_value_visible_on_instantiation():
    """Check if the full input value is rendered if the input is instantiated with it."""

    class MyApp(App):
        def compose(self):
            yield Input(value="value")

    app = MyApp()
    async with app.run_test():
        console = Console(width=5)
        with console.capture() as capture:
            console.print(app.query_one(Input).render())
        assert capture.get() == "value"


async def test_input_value_visible_after_value_assignment():
    """Check if the full input value is rendered if the value is assigned to programmatically."""

    class MyApp(App):
        def compose(self):
            yield Input()

        def on_mount(self):
            self.query_one(Input).value = "value"

    app = MyApp()
    async with app.run_test():
        console = Console(width=5)
        with console.capture() as capture:
            console.print(app.query_one(Input).render())
        assert capture.get() == "value"


async def test_input_value_visible_if_mounted_later():
    """Check if full input value is rendered if the widget is mounted later."""

    class MyApp(App):
        BINDINGS = [("a", "add_input", "add_input")]

        async def action_add_input(self):
            await self.mount(Input(value="value"))

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("a")
        await pilot.pause()
        console = Console(width=5)
        with console.capture() as capture:
            console.print(app.query_one(Input).render())
        assert capture.get() == "value"


async def test_input_value_visible_if_mounted_later_and_focused():
    """Check if full input value is rendered if the widget is mounted later and immediately focused."""

    class MyApp(App):
        BINDINGS = [("a", "add_input", "add_input")]

        async def action_add_input(self):
            inp = Input(value="value")
            await self.mount(inp)
            inp.focus()

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("a")
        await pilot.pause()
        console = Console(width=5)
        with console.capture() as capture:
            console.print(app.query_one(Input).render())
        assert capture.get() == "value"


async def test_input_value_visible_if_mounted_later_and_assigned_after():
    """Check if full value rendered if the widget is mounted later and the value is then assigned to."""

    class MyApp(App):
        BINDINGS = [
            ("a", "add_input", "add_input"),
            ("v", "set_value", "set_value"),
        ]

        async def action_add_input(self):
            await self.mount(Input())

        def action_set_value(self):
            self.query_one(Input).value = "value"

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.press("a")
        await pilot.press("v")
        console = Console(width=5)
        with console.capture() as capture:
            console.print(app.query_one(Input).render())
        assert capture.get() == "value"



from textual.widgets._markdown import (
    Markdown,
    MarkdownBlock,
    MarkdownTableOfContents,
    TableOfContentsType,
)

__all__ = [
    "MarkdownTableOfContents",
    "Markdown",
    "MarkdownBlock",
    "TableOfContentsType",
]



import pytest

from textual._binary_encode import DecodeError, dump, load


@pytest.mark.parametrize(
    "data",
    [
        None,
        False,
        True,
        -10,
        -1,
        0,
        1,
        100,
        "",
        "💩",
        "Hello",
        b"World",
        b"",
        [],
        (),
        [None],
        [1, 2, 3],
        (0, "foo"),
        (1, "foo"),
        (0, ""),
        ("", "💩", "💩💩"),
        (""),
        ["hello", "world"],
        ["hello", b"world"],
        ("hello", "world"),
        ("hello", b"world"),
        ("foo", "bar", "baz"),
        ("foo " * 1000, "bar " * 100, "baz " * 500),
        (1, "foo", "bar", "baz"),
        {},
        {"foo": "bar"},
        {"foo": "bar", b"egg": b"baz"},
        {"foo": "bar", b"egg": b"baz", "list_of_things": [1, 2, 3, "Paul", "Jessica"]},
        [{}],
        [[1]],
        [(1, 2), (3, 4)],
    ],
)
def test_round_trip(data: object) -> None:
    """Test the data may be encoded then decoded"""
    encoded = dump(data)
    assert isinstance(encoded, bytes)
    decoded = load(encoded)
    assert data == decoded


@pytest.mark.parametrize(
    "data",
    [
        b"",
        b"100:hello",
        b"i",
        b"i1",
        b"i10",
        b"li1e",
        b"x100",
    ],
)
def test_bad_encoding(data: bytes) -> None:
    with pytest.raises(DecodeError):
        load(data)


@pytest.mark.parametrize(
    "data",
    [
        set(),
        float,
        ...,
        [float],
    ],
)
def test_dump_invalid_type(data):
    with pytest.raises(TypeError):
        dump(data)


def test_load_wrong_type():
    with pytest.raises(TypeError):
        load(None)
    with pytest.raises(TypeError):
        load("foo")



from textual.app import App, ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Input


class Name(Widget):
    """Generates a greeting."""

    who = reactive("name", layout=True)  # (1)!

    def render(self) -> str:
        return f"Hello, {self.who}!"


class WatchApp(App):
    CSS_PATH = "refresh02.tcss"

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Enter your name")
        yield Name()

    def on_input_changed(self, event: Input.Changed) -> None:
        self.query_one(Name).who = event.value


if __name__ == "__main__":
    app = WatchApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Label
from textual.containers import Container


# Test fr dimensions and margins work in an auto container
# https://github.com/Textualize/textual/issues/2220
class TestApp(App):
    CSS = """
    Container {
        background: green 20%;
        border: heavy green;
        width: auto;
        height: auto;
        overflow: hidden;
    }

    Label {
        background: green 20%;      
        width: 1fr;
        height: 1fr;
        margin: 2 2;            
    }
    """

    def compose(self) -> ComposeResult:
        with Container():
            yield Label("Hello")
            yield Label("World")
            yield Label("!!")


if __name__ == "__main__":
    app = TestApp()
    app.run()



import re

import pytest

from textual.app import App, ComposeResult
from textual.widgets import Input
from textual.widgets._input import _RESTRICT_TYPES


def test_input_number_type():
    """Test number type regex, value should be number or the prefix of a valid number"""
    number = _RESTRICT_TYPES["number"]
    assert re.fullmatch(number, "0")
    assert re.fullmatch(number, "0.")
    assert re.fullmatch(number, ".")
    assert re.fullmatch(number, "-")
    assert re.fullmatch(number, "+")
    assert re.fullmatch(number, ".0")
    assert re.fullmatch(number, "1.1")
    assert re.fullmatch(number, "1_")
    assert re.fullmatch(number, "1_2")
    assert re.fullmatch(number, "-000_123_456.78e01_234")
    assert re.fullmatch(number, "1e1")
    assert re.fullmatch(number, "1")
    assert re.fullmatch(number, "1.")
    assert re.fullmatch(number, "1.2")
    assert re.fullmatch(number, "1.2e")
    assert re.fullmatch(number, "1.2e-")
    assert re.fullmatch(number, "1.2e-1")
    assert re.fullmatch(number, "1.2e-10")
    assert re.fullmatch(number, "1.2E10")
    assert not re.fullmatch(number, "1.2e10e")
    assert not re.fullmatch(number, "-000_123_456.78e01_234.")
    assert not re.fullmatch(number, "e")  # float("e23") is not valid
    assert not re.fullmatch(number, "1f2")
    assert not re.fullmatch(number, "inf")
    assert not re.fullmatch(number, "nan")
    assert not re.fullmatch(number, "-inf")



def test_input_integer_type():
    """Test input type regex"""
    integer = _RESTRICT_TYPES["integer"]
    assert re.fullmatch(integer, "0")
    assert re.fullmatch(integer, "1")
    assert re.fullmatch(integer, "10")
    assert re.fullmatch(integer, "123456789")
    assert re.fullmatch(integer, "-")
    assert re.fullmatch(integer, "+")
    assert re.fullmatch(integer, "-1")
    assert re.fullmatch(integer, "+2")
    assert re.fullmatch(integer, "+0")
    assert re.fullmatch(integer, "+0_")
    assert re.fullmatch(integer, "+0_1")
    assert re.fullmatch(integer, "+0_12")
    assert re.fullmatch(integer, "+0_123")
    assert not re.fullmatch(integer, "+_123")
    assert not re.fullmatch(integer, "123.")
    assert not re.fullmatch(integer, "+2e")
    assert not re.fullmatch(integer, "foo")


async def test_bad_type():
    """Check an invalid type raises a ValueError."""

    class InputApp(App):
        def compose(self) -> ComposeResult:
            yield Input(type="foo")  # Bad type

    app = InputApp()

    with pytest.raises(ValueError):
        async with app.run_test():
            pass


async def test_max_length():
    """Check max_length limits characters."""

    class InputApp(App):
        AUTO_FOCUS = "Input"

        def compose(self) -> ComposeResult:
            yield Input(max_length=5)

    async with InputApp().run_test() as pilot:
        input_widget = pilot.app.query_one(Input)
        await pilot.press("1")
        assert input_widget.value == "1"
        await pilot.press("2", "3", "4", "5")
        assert input_widget.value == "12345"
        # Value is max length, no more characters are permitted
        await pilot.press("6")
        assert input_widget.value == "12345"
        await pilot.press("7")
        assert input_widget.value == "12345"
        # Backspace is ok
        await pilot.press("backspace")
        assert input_widget.value == "1234"
        await pilot.press("0")
        assert input_widget.value == "12340"
        # Back to maximum
        await pilot.press("1")
        assert input_widget.value == "12340"


async def test_restrict():
    """Test restriction by regex."""

    class InputApp(App):
        AUTO_FOCUS = "Input"

        def compose(self) -> ComposeResult:
            yield Input(restrict="[abc]*")

    async with InputApp().run_test() as pilot:
        input_widget = pilot.app.query_one(Input)
        await pilot.press("a")
        assert input_widget.value == "a"
        await pilot.press("b")
        assert input_widget.value == "ab"
        await pilot.press("c")
        assert input_widget.value == "abc"
        # "d" is restricted
        await pilot.press("d")
        assert input_widget.value == "abc"
        # "a" is not
        await pilot.press("a")
        assert input_widget.value == "abca"


async def test_restrict_type():
    class InputApp(App):
        def compose(self) -> ComposeResult:
            yield Input(type="integer", id="integer")
            yield Input(type="number", id="number")
            yield Input(type="text", id="text")

    async with InputApp().run_test() as pilot:
        integer_input = pilot.app.query_one("#integer", Input)
        number_input = pilot.app.query_one("#number", Input)
        text_input = pilot.app.query_one("#text", Input)

        integer_input.focus()
        await pilot.press("a")
        assert not integer_input.value
        await pilot.press("-")
        assert integer_input.value == "-"
        assert integer_input.is_valid is False

        await pilot.press("1")
        assert integer_input.value == "-1"
        assert integer_input.is_valid is True

        number_input.focus()
        await pilot.press("x")
        assert number_input.value == ""
        await pilot.press("-", "3", ".", "1", "4", "y")
        assert number_input.value == "-3.14"

        text_input.focus()
        await pilot.press("!", "x", "9")
        assert text_input.value == "!x9"



from textual.app import App
from textual.containers import Horizontal, ScrollableContainer
from textual.widgets import Label

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain.
"""


class ScrollbarApp(App):
    CSS_PATH = "scrollbar_size2.tcss"

    def compose(self):
        yield Horizontal(
            ScrollableContainer(Label(TEXT * 5), id="v1"),
            ScrollableContainer(Label(TEXT * 5), id="v2"),
            ScrollableContainer(Label(TEXT * 5), id="v3"),
        )


if __name__ == "__main__":
    app = ScrollbarApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Button


class OnDecoratorApp(App):
    CSS_PATH = "on_decorator.tcss"

    def compose(self) -> ComposeResult:
        """Three buttons."""
        yield Button("Bell", id="bell")
        yield Button("Toggle dark", classes="toggle dark")
        yield Button("Quit", id="quit")

    def on_button_pressed(self, event: Button.Pressed) -> None:  # (1)!
        """Handle all button pressed events."""
        if event.button.id == "bell":
            self.bell()
        elif event.button.has_class("toggle", "dark"):
            self.dark = not self.dark
        elif event.button.id == "quit":
            self.exit()


if __name__ == "__main__":
    app = OnDecoratorApp()
    app.run()



from textual.widgets import Tree


def test_tree_node_parent() -> None:
    """It should be possible to access a TreeNode's parent."""
    tree = Tree[None]("Anakin")
    child = tree.root.add("Leia")
    grandchild = child.add("Ben")
    assert tree.root.parent is None
    assert grandchild.parent == child
    assert child.parent == tree.root
